#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import time

from dateutil import tz
from testtools import matchers

import yaql.tests

DT = datetime.datetime
TS = datetime.timedelta


class TestDatetime(yaql.tests.TestCase):
    def test_build_datetime_components(self):
        dt = DT(2015, 8, 29, tzinfo=tz.tzutc())
        self.assertEqual(
            dt, self.eval('datetime(2015, 8, 29)'))
        self.assertEqual(
            dt, self.eval('datetime(year => 2015, month => 8, day => 29,'
                          'hour => 0, minute => 0, second => 0, '
                          'microsecond => 0)'))

    def test_build_datetime_iso(self):
        self.assertEqual(
            DT(2015, 8, 29, tzinfo=tz.tzutc()),
            self.eval('datetime("2015-8-29")')
        )
        self.assertEqual(
            DT(2008, 9, 3, 20, 56, 35, 450686, tzinfo=tz.tzutc()),
            self.eval('datetime("2008-09-03T20:56:35.450686")')
        )
        self.assertEqual(
            DT(2008, 9, 3, 20, 56, 35, 450686, tzinfo=tz.tzutc()),
            self.eval('datetime("2008-09-03T20:56:35.450686Z")')
        )
        self.assertEqual(
            DT(2008, 9, 3, 0, 0, tzinfo=tz.tzutc()),
            self.eval('datetime("20080903")')
        )
        dt = self.eval('datetime("2008-09-03T20:56:35.450686+03:00")')
        self.assertEqual(
            DT(2008, 9, 3, 20, 56, 35, 450686),
            dt.replace(tzinfo=None)
        )
        self.assertEqual(TS(hours=3), dt.utcoffset())

    def test_build_datetime_string(self):
        self.assertEqual(
            DT(2006, 11, 21, 16, 30, tzinfo=tz.tzutc()),
            self.eval('datetime("Tuesday, 21. November 2006 04:30PM", '
                      '"%A, %d. %B %Y %I:%M%p")')
        )

    def test_datetime_fields(self):
        dt = DT(2006, 11, 21, 16, 30, tzinfo=tz.tzutc())
        self.assertEqual(2006, self.eval('$.year', dt))
        self.assertEqual(11, self.eval('$.month', dt))
        self.assertEqual(21, self.eval('$.day', dt))
        self.assertEqual(16, self.eval('$.hour', dt))
        self.assertEqual(30, self.eval('$.minute', dt))
        self.assertEqual(0, self.eval('$.second', dt))
        self.assertEqual(0, self.eval('$.microsecond', dt))
        self.assertEqual(1164126600, self.eval('$.timestamp', dt))
        self.assertEqual(1, self.eval('$.weekday', dt))
        self.assertEqual(TS(), self.eval('$.offset', dt))
        self.assertEqual(TS(hours=16, minutes=30), self.eval('$.time', dt))
        self.assertEqual(dt.replace(hour=0, minute=0), self.eval('$.date', dt))
        self.assertEqual(dt, self.eval('$.utc', dt))

    def test_build_timespan(self):
        self.assertEqual(TS(0), self.eval('timespan()'))
        self.assertEqual(
            TS(1, 7384, 5006),
            self.eval('timespan(days => 1, hours => 2, minutes => 3, '
                      'seconds => 4, milliseconds => 5, microseconds => 6)'))
        self.assertEqual(
            TS(1, 7384, 4994),
            self.eval('timespan(days => 1, hours => 2, minutes => 3, '
                      'seconds =>4, milliseconds => 5, microseconds => -6)'))

        self.assertEqual(
            TS(microseconds=-1000), self.eval('timespan(milliseconds => -1)'))

    def test_datetime_from_timestamp(self):
        dt = DT(2006, 11, 21, 16, 30, tzinfo=tz.tzutc())
        self.assertEqual(dt, self.eval('datetime(1164126600)'))

    def test_replace(self):
        dt = DT(2006, 11, 21, 16, 30, tzinfo=tz.tzutc())
        self.assertEqual(
            DT(2009, 11, 21, 16, 40, tzinfo=tz.tzutc()),
            self.eval('$.replace(year => 2009, minute => 40)', dt))

    def test_timespan_fields(self):
        ts = TS(1, 51945, 5000)
        self.assertAlmostEqual(1.6, self.eval('$.days', ts), places=2)
        self.assertAlmostEqual(38.43, self.eval('$.hours', ts), places=2)
        self.assertAlmostEqual(2305.75, self.eval('$.minutes', ts), places=2)
        self.assertAlmostEqual(138345, self.eval('$.seconds', ts), places=1)
        self.assertEqual(138345005, self.eval('$.milliseconds', ts))
        self.assertEqual(138345005000, self.eval('$.microseconds', ts))

    def test_now(self):
        self.assertIsInstance(self.eval('now()'), DT)
        self.assertIsInstance(self.eval('now(utctz())'), DT)
        self.assertIsInstance(self.eval('now(localtz())'), DT)
        self.assertThat(
            self.eval('now(utctz()) - now()'),
            matchers.LessThan(TS(seconds=1))
        )
        self.assertTrue(self.eval('now(localtz()).offset = localtz()'))

    def test_datetime_math(self):
        self.context['dt1'] = self.eval('now()')
        time.sleep(0.1)
        self.context['dt2'] = self.eval('now()')
        delta = TS(milliseconds=120)
        self.assertIsInstance(self.eval('$dt2 - $dt1'), TS)
        self.assertThat(self.eval('$dt2 - $dt1'), matchers.LessThan(delta))
        self.assertTrue(self.eval('($dt2 - $dt1) + $dt1 = $dt2'))
        self.assertTrue(self.eval('$dt1 + ($dt2 - $dt1) = $dt2'))
        self.assertThat(
            self.eval('($dt2 - $dt1) * 2'), matchers.LessThan(2 * delta))
        self.assertThat(
            self.eval('2.1 * ($dt2 - $dt1)'), matchers.LessThan(2 * delta))
        self.assertTrue(self.eval('-($dt1 - $dt2) = +($dt2 - $dt1)'))
        self.assertTrue(self.eval('$dt2 > $dt1'))
        self.assertTrue(self.eval('$dt2 >= $dt1'))
        self.assertTrue(self.eval('$dt2 != $dt1'))
        self.assertTrue(self.eval('$dt1 = $dt1'))
        self.assertTrue(self.eval('$dt1 < $dt2'))
        self.assertTrue(self.eval('$dt1 <= $dt2'))
        self.assertEqual(-1, self.eval('($dt2 - $dt1) / ($dt1 - $dt2)'))
        self.assertTrue(self.eval('$dt2 - ($dt2 - $dt1) = $dt1'))
        self.assertEqual(
            0, self.eval('($dt2 - $dt1) - ($dt2 - $dt1)').total_seconds())

        delta2 = self.eval('($dt2 - $dt1) / 2.1')
        self.assertThat(delta2, matchers.LessThan(delta / 2))
        self.assertTrue(self.eval('$dt1 + $ < $dt2', delta2))
        self.assertTrue(self.eval('$ + $dt1 < $dt2', delta2))
        self.assertTrue(self.eval('$dt2 - $dt1 > $', delta2))
        self.assertTrue(self.eval('$dt2 - $dt1 >= $', delta2))
        self.assertTrue(self.eval('$dt2 - $dt1 != $', delta2))
        self.assertFalse(self.eval('$dt2 - $dt1 < $', delta2))
        self.assertFalse(self.eval('$dt2 - $dt1 <= $', delta2))
        self.assertTrue(self.eval('($dt2 - $dt1) + $ > $', delta2))

    def test_is_datetime(self):
        self.assertTrue(self.eval('isDatetime(datetime("2015-8-29"))'))
        self.assertFalse(self.eval('isDatetime(123)'))
        self.assertFalse(self.eval('isDatetime(abc)'))

    def test_is_timespan(self):
        self.assertTrue(self.eval('isTimespan(timespan(milliseconds => -1))'))
        self.assertFalse(self.eval('isTimespan(123)'))
        self.assertFalse(self.eval('isTimespan(abc)'))



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Any Python class or object can be yaqlized. It is possible to call methods,
access attributes/properties and index of yaqlized objects.

The first way to yaqlize object is using function call:

.. code-block:: python

    class A(object):
        foo = 256
        def bar(self):
            print('yaqlization works with methods too')

    sample_object = A()
    yaqlization.yaqlize(sample_object)

The second way is using decorator:

.. code-block:: python

    @yaqlization.yaqlize
    class A(object):
        foo = 256
        def bar(self):
            print('yaqlization works with methods too')

Any mentioned operation on yaqlized objects can be disabled with additional
parameters for yaqlization. Also it is possible to specify whitelist/blacklist
of methods/attributes/keys that are exposed to the yaql.

This module provides implemented operators on Yaqlized objects.
"""


import re

from yaql.language import expressions
from yaql.language import runner
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes
from yaql import yaqlization


REGEX_TYPE = type(re.compile('.'))


class Yaqlized(yaqltypes.GenericType):
    def __init__(self, can_access_attributes=False, can_call_methods=False,
                 can_index=False):
        def check_value(value, context, *args, **kwargs):
            settings = yaqlization.get_yaqlization_settings(value)
            if settings is None:
                return False
            if can_access_attributes and not settings['yaqlizeAttributes']:
                return False
            if can_call_methods and not settings['yaqlizeMethods']:
                return False
            if can_index and not settings['yaqlizeIndexer']:
                return False
            return True

        super().__init__(checker=check_value, nullable=False)


def _match_name_to_entry(name, entry):
    if name == entry:
        return True
    elif isinstance(entry, REGEX_TYPE):
        return entry.search(name) is not None
    elif callable(entry):
        return entry(name)
    return False


def _validate_name(name, settings, exception_cls=AttributeError):
    if name.startswith('_'):
        raise exception_cls('Cannot access ' + name)
    whitelist = settings['whitelist']
    if whitelist:
        for entry in whitelist:
            if _match_name_to_entry(name, entry):
                return
        raise exception_cls('Cannot access ' + name)
    blacklist = settings['blacklist']
    if blacklist:
        for entry in blacklist:
            if _match_name_to_entry(name, entry):
                raise exception_cls('Cannot access ' + name)


def _remap_name(name, settings):
    return settings['attributeRemapping'].get(name, name)


def _auto_yaqlize(value, settings):
    if not settings['autoYaqlizeResult']:
        return
    if isinstance(value, type):
        cls = value
    else:
        cls = type(value)
    if cls.__module__ == int.__module__:
        return
    try:
        yaqlization.yaqlize(value, auto_yaqlize_result=True)
    except Exception:
        pass


@specs.parameter('receiver', Yaqlized(can_call_methods=True))
@specs.parameter('expr', yaqltypes.YaqlExpression(expressions.Function))
@specs.name('#operator_.')
def op_dot(receiver, expr, context, engine):
    """:yaql:operator .

    Evaluates expression on receiver and returns its result.

    :signature: receiver.expr
    :arg receiver: yaqlized receiver
    :argType receiver: yaqlized object, initialized with
        yaqlize_methods equal to True
    :arg expr: expression to be evaluated
    :argType expr: expression
    :returnType: any (expression return type)
    """
    settings = yaqlization.get_yaqlization_settings(receiver)
    mappings = _remap_name(expr.name, settings)

    _validate_name(expr.name, settings)
    if not isinstance(mappings, str):
        name = mappings[0]
        if len(mappings) > 0:
            arg_mappings = mappings[1]
        else:
            arg_mappings = {}
    else:
        name = mappings
        arg_mappings = {}

    func = getattr(receiver, name)
    args, kwargs = runner.translate_args(False, expr.args, {})
    args = tuple(arg(utils.NO_VALUE, context, engine) for arg in args)
    for key, value in kwargs.items():
        kwargs[arg_mappings.get(key, key)] = value(
            utils.NO_VALUE, context, engine)
    res = func(*args, **kwargs)
    _auto_yaqlize(res, settings)
    return res


@specs.parameter('obj', Yaqlized(can_access_attributes=True))
@specs.parameter('attr', yaqltypes.Keyword())
@specs.name('#operator_.')
def attribution(obj, attr):
    """:yaql:operator .

    Returns attribute of the object.

    :signature: obj.attr
    :arg obj: yaqlized object
    :argType obj: yaqlized object, initialized with
        yaqlize_attributes equal to True
    :arg attr: attribute name
    :argType attr: keyword
    :returnType: any
    """
    settings = yaqlization.get_yaqlization_settings(obj)
    _validate_name(attr, settings)
    attr = _remap_name(attr, settings)
    res = getattr(obj, attr)
    _auto_yaqlize(res, settings)
    return res


@specs.parameter('obj', Yaqlized(can_index=True))
@specs.name('#indexer')
def indexation(obj, key):
    """:yaql:operator indexer

    Returns value of attribute/property key of the object.

    :signature: obj[key]
    :arg obj: yaqlized object
    :argType obj: yaqlized object, initialized with
        yaqlize_indexer equal to True
    :arg key: index name
    :argType key: keyword
    :returnType: any
    """
    settings = yaqlization.get_yaqlization_settings(obj)
    _validate_name(key, settings, KeyError)
    res = obj[key]
    _auto_yaqlize(res, settings)
    return res


def register(context):
    context = context.create_child_context()
    context.register_function(op_dot)
    context.register_function(attribution)
    context.register_function(indexation)
    return context



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
The module describes main system functions for working with objects.
"""

import itertools

from yaql.language import contexts
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes


@specs.parameter('name', yaqltypes.StringConstant())
@specs.name('#get_context_data')
def get_context_data(name, context):
    """:yaql:getContextData

    Returns the context value by its name. This function is system
    and can be overridden to change the way of getting context data.

    :signature: getContextData(name)
    :arg name: value's key name
    :argType name: string
    :returnType: any (value type)
    """
    return context[name]


@specs.parameter('expr', yaqltypes.Lambda(method=True))
@specs.name('#operator_.')
def op_dot(receiver, expr):
    """:yaql:operator .

    Returns expr evaluated on receiver.

    :signature: receiver.expr
    :arg receiver: object to evaluate expression
    :argType receiver: any
    :arg expr: expression
    :argType expr: expression that can be evaluated as a method
    :returnType: expression result type

    .. code::

        yaql> [0, 1].select($+1)
        [1, 2]
    """
    return expr(receiver)


@specs.parameter('expr', yaqltypes.YaqlExpression())
@specs.inject('operator', yaqltypes.Delegate('#operator_.'))
@specs.name('#operator_?.')
def elvis_operator(operator, receiver, expr):
    """:yaql:operator ?.

    Evaluates expr on receiver if receiver isn't null and returns the result.
    If receiver is null returns null.

    :signature: receiver?.expr
    :arg receiver: object to evaluate expression
    :argType receiver: any
    :arg expr: expression
    :argType expr: expression that can be evaluated as a method
    :returnType: expression result or null

    .. code::

        yaql> [0, 1]?.select($+1)
        [1, 2]
        yaql> null?.select($+1)
        null
    """
    if receiver is None:
        return None
    return operator(receiver, expr)


@specs.parameter('sequence', yaqltypes.Iterable())
@specs.parameter('args', yaqltypes.String())
@specs.method
def unpack(sequence, context, *args):
    """:yaql:unpack

    Returns context object with sequence values unpacked to args.
    If args size is equal to sequence size then args get appropriate
    sequence values.
    If args size is 0 then args are 1-based indexes.
    Otherwise ValueError is raised.

    :signature: sequence.unpack([args])
    :receiverArg sequence: iterable of items to be passed as context values
    :argType sequence: iterable
    :arg [args]: keys to be associated with sequence items. If args size is
        equal to sequence size then args get appropriate sequence items. If
        args size is 0 then args are indexed start from 1. Otherwise exception
        will be thrown
    :argType [args]: chain of strings
    :returnType: context object

    .. code::

        yaql> [1, 2].unpack(a, b) -> $a + $b
        3
        yaql> [2, 3].unpack() -> $1 + $2
        5
    """
    lst = tuple(itertools.islice(sequence, len(args) + 1))
    if 0 < len(args) != len(lst):
        raise ValueError('Cannot unpack {} elements into {}'.format(
            len(lst), len(args)))
    if len(args) > 0:
        for i in range(len(lst)):
            context[args[i]] = lst[i]
    else:
        for i, t in enumerate(sequence, 1):
            context[str(i)] = t
    return context


def with_(context, *args):
    """:yaql:with

    Returns new context object where args are stored with 1-based indexes.

    :signature: with([args])
    :arg [args]: values to be stored under appropriate numbers $1, $2, ...
    :argType [args]: chain of any values
    :returnType: context object

    .. code::

        yaql> with("ab", "cd") -> $1 + $2
        "abcd"
    """
    for i, t in enumerate(args, 1):
        context[str(i)] = t
    return context


@specs.inject('__context__', yaqltypes.Context())
def let(__context__, *args, **kwargs):
    """:yaql:let

    Returns context object where args are stored with 1-based indexes
    and kwargs values are stored with appropriate keys.

    :signature: let([args], {kwargs})
    :arg [args]: values to be stored under appropriate numbers $1, $2, ...
    :argType [args]: chain of any values
    :arg {kwargs}: values to be stored under appropriate keys
    :argType {kwargs}: chain of mappings
    :returnType: context object

    .. code::

        yaql> let(1, 2, a => 3, b => 4) -> $1 + $a + $2 + $b
        10
    """
    for i, value in enumerate(args, 1):
        __context__[str(i)] = value

    for key, value in kwargs.items():
        __context__[key] = value
    return __context__


@specs.parameter('name', yaqltypes.String())
@specs.parameter('func', yaqltypes.Lambda())
def def_(name, func, context):
    """:yaql:def

    Returns new context object with function name defined.

    :signature: def(name, func)
    :arg name: name of function
    :argType name: string
    :arg func: function to be stored under provided name
    :argType func: lambda
    :returnType: context object

    .. code::

        yaql> def(sq, $*$) -> [1, 2, 3].select(sq($))
        [1, 4, 9]
    """
    @specs.name(name)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    context.register_function(wrapper)
    return context


@specs.parameter('left', contexts.ContextBase)
@specs.parameter('right', yaqltypes.Lambda(with_context=True))
@specs.name('#operator_->')
def send_context(left, right):
    """:yaql:operator ->

    Evaluates lambda on provided context and returns the result.

    :signature: left -> right
    :arg left: context to be used for function
    :argType left: context object
    :arg right: function
    :argType right: lambda
    :returnType: any (function return value type)

    .. code::

        yaql> let(a => 1) -> $a
        1
    """
    return right(left)


@specs.method
@specs.parameter('condition', yaqltypes.Lambda())
@specs.parameter('message', yaqltypes.String())
def assert__(engine, obj, condition, message='Assertion failed'):
    """:yaql:assert

    Evaluates condition against object. If it evaluates to true returns the
    object, otherwise throws an exception with provided message.

    :signature: obj.assert(condition, message => "Assertion failed")
    :arg obj: object to evaluate condition on
    :argType obj: any
    :arg condition: lambda function to be evaluated on obj. If result of
        function evaluates to false then trows exception message
    :argType condition: lambda
    :arg message: message to trow if condition returns false
    :argType message: string
    :returnType: obj type or message

    .. code::

        yaql> 12.assert($ < 2)
        Execution exception: Assertion failed
        yaql> 12.assert($ < 20)
        12
        yaql> [].assert($, "Failed assertion")
        Execution exception: Failed assertion
    """
    if utils.is_iterator(obj):
        obj = utils.memorize(obj, engine)
    if not condition(obj):
        raise AssertionError(message)
    return obj


@specs.name('#call')
@specs.parameter('callable_', yaqltypes.PythonType(
    object, False, validators=(callable,)))
def call(callable_, *args, **kwargs):
    """:yaql:call

    Evaluates function with specified args and kwargs and returns the
    result.
    This function is used to transform expressions like '$foo(args, kwargs)'
    to '#call($foo, args, kwargs)'.
    Note that to use this functionality 'delegate' mode has to be enabled.

    :signature: call(callable, args, kwargs)
    :arg callable: callable function
    :argType callable: python type
    :arg args: sequence of items to be used for calling
    :argType args: sequence
    :arg kwargs: dictionary with kwargs to be used for calling
    :argType kwargs: mapping
    :returnType: any (callable return type)
    """
    return callable_(*args, **kwargs)


@specs.parameter('func', yaqltypes.Lambda())
def lambda_(func):
    """:yaql:lambda

    Constructs a new anonymous function
    Note that to use this function 'delegate' mode has to be enabled.

    :signature: lambda(func)
    :arg func: function to be returned
    :argType func: lambda
    :returnType: obj type or message

    .. code::

        yaql> let(func => lambda(2 * $)) -> [1, 2, 3].select($func($))
        [2, 4, 6]
        yaql> [1, 2, 3, 4].where(lambda($ > 3)($ + 1))
        [3, 4]
    """
    return func


@specs.name('#operator_.')
@specs.parameter('name', yaqltypes.Keyword())
@specs.inject('func', yaqltypes.Delegate(use_convention=False))
def get_property(func, obj, name):
    """:yaql:operator .

    Returns value of 'name' property.

    :signature: left.right
    :arg left: object
    :argType left: any
    :arg right: object property name
    :argType right: keyword
    :returnType: any

    .. code::

        yaql> now().year
        2016
    """
    func_name = '#property#{}'.format(name)
    return func(func_name, obj)


@specs.name('call')
@specs.parameter('name', yaqltypes.String())
@specs.parameter('args', yaqltypes.Sequence())
@specs.parameter('kwargs', utils.MappingType)
def call_func(context, engine, name, args, kwargs, receiver=utils.NO_VALUE):
    """:yaql:call

    Evaluates function name with specified args and kwargs and returns the
    result.

    :signature: call(name, args, kwargs)
    :arg name: name of callable
    :argType name: string
    :arg args: sequence of items to be used for calling
    :argType args: sequence
    :arg kwargs: dictionary with kwargs to be used for calling
    :argType kwargs: mapping
    :returnType: any (callable return type)

    .. code::

        yaql> call(let, [1, 2], {a => 3, b => 4}) -> $1 + $a + $2 + $b
        10
    """
    return context(name, engine, receiver)(
        *args, **utils.filter_parameters_dict(kwargs))


def register(context, delegates=False):
    context.register_function(get_context_data)
    context.register_function(op_dot)
    context.register_function(unpack)
    context.register_function(with_)
    context.register_function(send_context)
    context.register_function(let)
    context.register_function(def_)
    context.register_function(elvis_operator)
    context.register_function(assert__)
    context.register_function(call_func)
    if delegates:
        context.register_function(call)
        context.register_function(lambda_)


def register_fallbacks(context):
    context.register_function(get_property)



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from yaql.language import exceptions
from yaql.language import runner
from yaql.language import specs
from yaql.language import utils


class ContextBase(metaclass=abc.ABCMeta):
    def __init__(self, parent_context=None, convention=None):
        self._parent_context = parent_context
        self._convention = convention
        if convention is None and parent_context:
            self._convention = parent_context.convention

    @property
    def parent(self):
        return self._parent_context

    @abc.abstractmethod
    def register_function(self, spec, *args, **kwargs):
        pass

    @abc.abstractmethod
    def get_data(self, name, default=None, ask_parent=True):
        return default

    def __getitem__(self, name):
        return self.get_data(name)

    @abc.abstractmethod
    def __setitem__(self, name, value):
        pass

    @abc.abstractmethod
    def __delitem__(self, name):
        pass

    @abc.abstractmethod
    def __contains__(self, item):
        return False

    def __call__(self, name, engine, receiver=utils.NO_VALUE,
                 data_context=None, use_convention=False,
                 function_filter=None):
        return lambda *args, **kwargs: runner.call(
            name, self, args, kwargs, engine, receiver,
            data_context, use_convention, function_filter)

    @abc.abstractmethod
    def get_functions(self, name, predicate=None, use_convention=False):
        return [], False

    @abc.abstractmethod
    def delete_function(self, spec):
        pass

    def collect_functions(self, name, predicate=None, use_convention=False):
        overloads = []
        p = self
        while p is not None:
            context_predicate = None
            if predicate:
                context_predicate = lambda fd: predicate(fd, p)  # noqa: E731
            layer_overloads, is_exclusive = p.get_functions(
                name, context_predicate, use_convention)
            p = None if is_exclusive else p.parent
            if layer_overloads:
                overloads.append(layer_overloads)
        return overloads

    def create_child_context(self):
        return type(self)(self)

    @property
    def convention(self):
        return self._convention

    @abc.abstractmethod
    def keys(self):
        return {}.keys()


class Context(ContextBase):
    def __init__(self, parent_context=None, data=utils.NO_VALUE,
                 convention=None):
        super().__init__(parent_context, convention)
        self._functions = {}
        self._data = {}
        self._exclusive_funcs = set()
        if data is not utils.NO_VALUE:
            self['$'] = data

    @staticmethod
    def _import_function_definition(fd):
        return fd

    def register_function(self, spec, *args, **kwargs):
        exclusive = kwargs.pop('exclusive', False)

        if not isinstance(spec, specs.FunctionDefinition) and callable(spec):
            spec = specs.get_function_definition(
                spec, *args, convention=self._convention, **kwargs)

        spec = self._import_function_definition(spec)
        if spec.is_method:
            if not spec.is_valid_method():
                raise exceptions.InvalidMethodException(spec.name)
        self._functions.setdefault(spec.name, set()).add(spec)
        if exclusive:
            self._exclusive_funcs.add(spec.name)

    def delete_function(self, spec):
        self._functions.get(spec.name, set()).discard(spec)
        self._exclusive_funcs.discard(spec.name)

    def get_functions(self, name, predicate=None, use_convention=False):
        name = name.rstrip('_')
        if use_convention and self._convention is not None:
            name = self._convention.convert_function_name(name)
        if predicate is None:
            predicate = lambda x: True  # noqa: E731
        return (
            set(filter(predicate, self._functions.get(name, set()))),
            name in self._exclusive_funcs
        )

    @staticmethod
    def _normalize_name(name):
        if not name.startswith('$'):
            name = ('$' + name)
        if name == '$':
            name = '$1'
        return name

    def __setitem__(self, name, value):
        self._data[self._normalize_name(name)] = value

    def get_data(self, name, default=None, ask_parent=True):
        name = self._normalize_name(name)
        if name in self._data:
            return self._data[name]
        ctx = self.parent
        while ask_parent and ctx:
            result = ctx.get_data(name, utils.NO_VALUE, False)
            if result is utils.NO_VALUE:
                ctx = ctx.parent
            else:
                return result
        return default

    def __delitem__(self, name):
        self._data.pop(self._normalize_name(name))

    def __contains__(self, item):
        if isinstance(item, specs.FunctionDefinition):
            return item in self._functions.get(item.name, [])
        if isinstance(item, str):
            return self._normalize_name(item) in self._data
        return False

    def keys(self):
        return self._data.keys()


class MultiContext(ContextBase):
    def __init__(self, context_list, convention=None):
        self._context_list = context_list
        if convention is None:
            convention = context_list[0].convention
        parents = tuple(
            filter(lambda t: t, map(lambda t: t.parent, context_list))
        )
        if not parents:
            super().__init__(None, convention)
        elif len(parents) == 1:
            super().__init__(parents[0], convention)
        else:
            super().__init__(MultiContext(parents),
                             convention)

    def register_function(self, spec, *args, **kwargs):
        self._context_list[0].register_function(spec, *args, **kwargs)

    def get_data(self, name, default=None, ask_parent=True):
        for context in self._context_list:
            result = context.get_data(name, utils.NO_VALUE, False)
            if result is not utils.NO_VALUE:
                return result
        ctx = self.parent
        while ask_parent and ctx:
            result = ctx.get_data(name, utils.NO_VALUE, False)
            if result is utils.NO_VALUE:
                ctx = ctx.parent
            else:
                return result
        return default

    def __setitem__(self, name, value):
        self._context_list[0][name] = value

    def __delitem__(self, name):
        for context in self._context_list:
            del context[name]

    def create_child_context(self):
        return Context(self)

    def keys(self):
        prev_keys = set()
        for context in self._context_list:
            for key in context.keys():
                if key not in prev_keys:
                    prev_keys.add(key)
                    yield key

    def delete_function(self, spec):
        for context in self._context_list:
            context.delete_function(spec)

    def __contains__(self, item):
        for context in self._context_list:
            if item in context:
                return True
        return False

    def get_functions(self, name, predicate=None, use_convention=False):
        result = set()
        is_exclusive = False
        for context in self._context_list:
            funcs, exclusive = context.get_functions(
                name, predicate, use_convention)
            result.update(funcs)
            if exclusive:
                is_exclusive = True
        return result, is_exclusive


class LinkedContext(ContextBase):
    """Context that is as a proxy to another context but has its own parent."""

    def __init__(self, parent_context, linked_context, convention=None):
        self.linked_context = linked_context
        if linked_context.parent:
            super().__init__(
                LinkedContext(parent_context, linked_context.parent,
                              convention), convention)
        else:
            super().__init__(parent_context, convention)

    def register_function(self, spec, *args, **kwargs):
        return self.linked_context.register_function(spec, *args, **kwargs)

    def keys(self):
        return self.linked_context.keys()

    def get_data(self, name, default=None, ask_parent=True):
        result = self.linked_context.get_data(
            name, default=utils.NO_VALUE, ask_parent=False)
        if result is utils.NO_VALUE:
            if not ask_parent or not self.parent:
                return default
            return self.parent.get_data(name, default=default, ask_parent=True)
        return result

    def get_functions(self, name, predicate=None, use_convention=False):
        return self.linked_context.get_functions(
            name, predicate=predicate, use_convention=use_convention)

    def delete_function(self, spec):
        return self.linked_context.delete_function(spec)

    def __contains__(self, item):
        return item in self.linked_context

    def __delitem__(self, name):
        del self.linked_context[name]

    def __setitem__(self, name, value):
        self.linked_context[name] = value

    def create_child_context(self):
        return type(self.linked_context)(self)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import exceptions
import yaql.tests


class TestQueries(yaql.tests.TestCase):
    def test_where(self):
        data = [1, 2, 3, 4, 5, 6]
        self.assertEqual([4, 5, 6], self.eval('$.where($ > 3)', data=data))

    def test_select(self):
        data = [1, 2, 3]
        self.assertEqual([1, 4, 9], self.eval('$.select($ * $)', data=data))

    def test_keyword_collection_access(self):
        data = [{'a': 2}, {'a': 4}]
        self.assertEqual([2, 4], self.eval('$.a', data=data))
        self.assertEqual([2, 4], self.eval('$.select($).a', data=data))

    def test_skip(self):
        data = [1, 2, 3, 4]
        self.assertEqual([2, 3, 4], self.eval('$.skip(1)', data=data))

    def test_limit(self):
        data = [1, 2, 3, 4]
        self.assertEqual([1, 2], self.eval('$.limit(2)', data=data))
        self.assertEqual([1, 2], self.eval('$.take(2)', data=data))

    def test_append(self):
        data = [1, 2]
        self.assertEqual([1, 2, 3, 4], self.eval('$.append(3, 4)', data=data))

    def test_complex_query(self):
        data = [1, 2, 3, 4, 5, 6]
        self.assertEqual(
            [4],
            self.eval('$.where($ < 4).select($ * $).skip(1).limit(1)',
                      data=data))

    def test_distinct(self):
        data = [1, 2, 3, 2, 4, 8]
        self.assertEqual([1, 2, 3, 4, 8], self.eval('$.distinct()', data=data))
        self.assertEqual([1, 2, 3, 4, 8], self.eval('distinct($)', data=data))

    def test_distinct_structures(self):
        data = [{'a': 1}, {'b': 2}, {'a': 1}]
        self.assertEqual(
            [{'a': 1}, {'b': 2}],
            self.eval('$.distinct()', data=data))

    def test_distinct_with_selector(self):
        data = [['a', 1], ['b', 2], ['c', 1], ['d', 3], ['e', 2]]
        self.assertCountEqual([['a', 1], ['b', 2], ['d', 3]],
                              self.eval('$.distinct($[1])', data=data))
        self.assertCountEqual([['a', 1], ['b', 2], ['d', 3]],
                              self.eval('distinct($, $[1])', data=data))

    def test_any(self):
        self.assertFalse(self.eval('$.any()', data=[]))
        self.assertTrue(self.eval('$.any()', data=[0]))

    def test_all(self):
        self.assertTrue(self.eval('$.all()', data=[]))
        self.assertFalse(self.eval('$.all()', data=[1, 0]))
        self.assertTrue(self.eval('$.all()', data=[1, 2]))
        self.assertFalse(self.eval('$.all($ > 1)', data=[2, 1]))
        self.assertTrue(self.eval('$.all($ > 1)', data=[2, 3]))

    def test_enumerate(self):
        data = [1, 2, 3]
        self.assertEqual([[0, 1], [1, 2], [2, 3]],
                         self.eval('$.enumerate()', data=data))
        self.assertEqual([[3, 1], [4, 2], [5, 3]],
                         self.eval('$.enumerate(3)', data=data))
        self.assertEqual([[0, 1], [1, 2], [2, 3]],
                         self.eval('enumerate($)', data=data))
        self.assertEqual([[3, 1], [4, 2], [5, 3]],
                         self.eval('enumerate($, 3)', data=data))

    def test_concat(self):
        data = [1, 2, 3]
        self.assertEqual(
            [1, 2, 3, 2, 4, 6],
            self.eval('$.select($).concat($.select(2 * $))', data=data))
        self.assertEqual(
            [1, 2, 3, 2, 4, 6, 1, 2, 3],
            self.eval('concat($, $.select(2 * $), $)', data=data))

    def test_len(self):
        data = [1, 2, 3]
        self.assertEqual(3, self.eval('len($)', data=data))
        self.assertEqual(3, self.eval('$.len()', data=data))
        self.assertEqual(3, self.eval('$.count()', data=data))
        self.assertRaises(
            exceptions.FunctionResolutionError,
            self.eval, 'count($)', data=data)

    def test_sum(self):
        data = range(4)
        self.assertEqual(6, self.eval('$.sum()', data=data))
        self.assertEqual(106, self.eval('$.sum(100)', data=data))
        self.assertEqual(100, self.eval('[].sum(100)'))

    def test_memorize(self):
        generator_func = lambda: (i for i in range(3))  # noqa: E731
        self.assertRaises(
            TypeError,
            self.eval, '$.len() + $.sum()', data=generator_func())

        self.assertEqual(
            6,
            self.eval('let($.memorize()) -> $.len() + $.sum()',
                      data=generator_func()))

    def test_first(self):
        self.assertEqual(2, self.eval('list(2, 3).first()'))
        self.assertEqual(4, self.eval('list(2, 3).select($ * 2).first()'))
        self.assertIsNone(self.eval('list().first(null)'))
        self.assertRaises(StopIteration, self.eval, 'list().first()')
        self.assertEqual(99, self.eval('list().first(99)'))

    def test_single(self):
        self.assertEqual(2, self.eval('list(2).single()'))
        self.assertRaises(StopIteration, self.eval, 'list().single()')
        self.assertRaises(StopIteration, self.eval, 'list(1, 2).single()')

    def test_last(self):
        self.assertEqual(3, self.eval('list(2, 3).last()'))
        self.assertEqual(6, self.eval('list(2, 3).select($ * 2).last()'))
        self.assertIsNone(self.eval('list().last(null)'))
        self.assertEqual(99, self.eval('list().last(99)'))
        self.assertRaises(StopIteration, self.eval, 'list().last()')

    def test_range(self):
        self.assertEqual([0, 1], self.eval('range(2)'))
        self.assertEqual([1, 2, 3], self.eval('range(1, 4)'))
        self.assertEqual([4, 3, 2], self.eval('range(4, 1, -1)'))

    def test_select_many(self):
        self.assertEqual([0, 0, 1, 0, 1, 2],
                         self.eval('range(4).selectMany(range($))'))

    def test_select_many_scalar(self):
        # check that string is not interpreted as a sequence and that
        # selectMany works when selector returns scalar
        self.assertEqual(
            ['xx', 'xx'],
            self.eval('range(2).selectMany(xx)'))

    def test_order_by(self):
        self.assertEqual(
            [1, 2, 3, 4],
            self.eval('$.orderBy($)', data=[4, 2, 1, 3]))

        self.assertEqual(
            [4, 3, 2, 1],
            self.eval('$.orderByDescending($)', data=[4, 2, 1, 3]))

    def test_order_by_multilevel(self):
        self.assertEqual(
            [[1, 0], [1, 5], [2, 2]],
            self.eval(
                '$.orderBy($[0]).thenBy($[1])',
                data=[[2, 2], [1, 5], [1, 0]]))

        self.assertEqual(
            [[1, 5], [1, 0], [2, 2]],
            self.eval(
                '$.orderBy($[0]).thenByDescending($[1])',
                data=[[2, 2], [1, 5], [1, 0]]))

        self.assertEqual(
            [[2, 2], [1, 0], [1, 5]],
            self.eval(
                '$.orderByDescending($[0]).thenBy($[1])',
                data=[[2, 2], [1, 5], [1, 0]]))

        self.assertEqual(
            [[2, 2], [1, 5], [1, 0]],
            self.eval(
                '$.orderByDescending($[0]).thenByDescending($[1])',
                data=[[2, 2], [1, 5], [1, 0]]))

    def test_group_by(self):
        data = {'a': 1, 'b': 2, 'c': 1, 'd': 3, 'e': 2}
        self.assertCountEqual(
            [
                [1, [['a', 1], ['c', 1]]],
                [2, [['b', 2], ['e', 2]]],
                [3, [['d', 3]]]
            ],
            self.eval('$.items().orderBy($[0]).groupBy($[1])', data=data))

        self.assertCountEqual(
            [[1, ['a', 'c']], [2, ['b', 'e']], [3, ['d']]],
            self.eval('$.items().orderBy($[0]).groupBy($[1], $[0])',
                      data=data))

        self.assertCountEqual(
            [[1, 'ac'], [2, 'be'], [3, 'd']],
            self.eval('$.items().orderBy($[0]).'
                      'groupBy($[1], $[0], $.sum())', data=data))

        self.assertCountEqual(
            [[1, ['a', 1, 'c', 1]], [2, ['b', 2, 'e', 2]], [3, ['d', 3]]],
            self.eval('$.items().orderBy($[0]).'
                      'groupBy($[1],,  $.sum())',
                      data=data))

        self.assertCountEqual(
            [[1, ['a', 1, 'c', 1]], [2, ['b', 2, 'e', 2]], [3, ['d', 3]]],
            self.eval('$.items().orderBy($[0]).'
                      'groupBy($[1], aggregator => $.sum())',
                      data=data))

    def test_group_by_old_syntax(self):
        # Test the syntax used in 1.1.1 and earlier, where the aggregator
        # function was passed the key as well as the value list, and returned
        # the key along with the aggregated value. This ensures backward
        # compatibility with existing expressions.
        data = {'a': 1, 'b': 2, 'c': 1, 'd': 3, 'e': 2}

        self.assertCountEqual(
            [[1, 'ac'], [2, 'be'], [3, 'd']],
            self.eval('$.items().orderBy($[0]).'
                      'groupBy($[1], $[0], [$[0], $[1].sum()])', data=data))

        self.assertCountEqual(
            [[1, ['a', 1, 'c', 1]], [2, ['b', 2, 'e', 2]], [3, ['d', 3]]],
            self.eval('$.items().orderBy($[0]).'
                      'groupBy($[1],,  [$[0], $[1].sum()])',
                      data=data))

        self.assertCountEqual(
            [[1, ['a', 1, 'c', 1]], [2, ['b', 2, 'e', 2]], [3, ['d', 3]]],
            self.eval('$.items().orderBy($[0]).'
                      'groupBy($[1], aggregator => [$[0], $[1].sum()])',
                      data=data))

    def test_join(self):
        self.assertEqual(
            [[2, 1], [3, 1], [3, 2], [4, 1], [4, 2], [4, 3]],
            self.eval('$.join($, $1 > $2, [$1, $2])', data=[1, 2, 3, 4]))

        self.assertEqual(
            [[1, 3], [1, 4], [2, 3], [2, 4]],
            self.eval('[1,2].join([3, 4], true, [$1, $2])'))

    def test_zip(self):
        self.assertEqual(
            [[1, 4], [2, 5]],
            self.eval('[1, 2, 3].zip([4, 5])'))

        self.assertEqual(
            [[1, 4, 6], [2, 5, 7]],
            self.eval('[1, 2, 3].zip([4, 5], [6, 7, 8])'))

    def test_zip_longest(self):
        self.assertEqual(
            [[1, 4], [2, 5], [3, None]],
            self.eval('[1, 2, 3].zipLongest([4, 5])'))

        self.assertEqual(
            [[1, 4, 6], [2, 5, None], [3, None, None]],
            self.eval('[1, 2, 3].zipLongest([4, 5], [6])'))

        self.assertEqual(
            [[1, 4], [2, 5], [3, 0]],
            self.eval('[1, 2, 3].zipLongest([4, 5], default => 0)'))

    def test_repeat(self):
        self.assertEqual(
            [None, None],
            self.eval('null.repeat(2)'))

        self.assertEqual(
            [1, 1, 1, 1, 1],
            self.eval('1.repeat().limit(5)'))

    def test_cycle(self):
        self.assertEqual(
            [1, 2, 1, 2, 1],
            self.eval('[1, 2].cycle().take(5)'))

    def test_take_while(self):
        self.assertEqual(
            [1, 2, 3],
            self.eval('[1, 2, 3, 4, 5].takeWhile($ < 4)'))

    def test_skip_while(self):
        self.assertEqual(
            [4, 5],
            self.eval('[1, 2, 3, 4, 5].skipWhile($ < 4)'))

    def test_index_of(self):
        self.assertEqual(1, self.eval('[1, 2, 3, 2, 1].indexOf(2)'))
        self.assertEqual(-1, self.eval('[1, 2, 3, 2, 1].indexOf(22)'))

    def test_last_index_of(self):
        self.assertEqual(3, self.eval('[1, 2, 3, 2, 1].lastIndexOf(2)'))
        self.assertEqual(-1, self.eval('[1, 2, 3, 2, 1].lastIndexOf(22)'))

    def test_index_where(self):
        self.assertEqual(1, self.eval('[1, 2, 3, 2, 1].indexWhere($ = 2)'))
        self.assertEqual(-1, self.eval('[1, 2, 3, 2, 1].indexWhere($ = 22)'))

    def test_last_index_where(self):
        self.assertEqual(3, self.eval('[1, 2, 3, 2, 1].lastIndexWhere($ = 2)'))
        self.assertEqual(
            -1, self.eval('[1, 2, 3, 2, 1].lastIndexWhere($ = 22)'))

    def test_slice(self):
        self.assertEqual(
            [[1, 2], [3, 4], [5]],
            self.eval('range(1, 6).slice(2)'))
        self.assertEqual(
            [[1, 2], [3, 4], [5]],
            self.eval('[1,2,3,4,5].slice(2)'))

    def test_split_where(self):
        self.assertEqual(
            [[], [2, 3], [5]],
            self.eval('range(1, 6).splitWhere($ mod 3 = 1)'))

    def test_split_at(self):
        self.assertEqual(
            [[1, 2], [3, 4, 5]],
            self.eval('range(1, 6).splitAt(2)'))

    def test_slice_where(self):
        self.assertEqual(
            [['a', 'a'], ['b'], ['a', 'a']],
            self.eval('[a,a,b,a,a].sliceWhere($ != a)'))

    def test_aggregate(self):
        self.assertEqual(
            'aabaa',
            self.eval('[a,a,b,a,a].aggregate($1 + $2)'))

        self.assertRaises(
            TypeError,
            self.eval, '[].aggregate($1 + $2)')

        self.assertEqual(
            1,
            self.eval('[].aggregate($1 + $2, 1)'))

        self.assertEqual(
            'aabaa',
            self.eval('[a,a,b,a,a].reduce($1 + $2)'))

        self.assertEqual(
            0,
            self.eval('[].reduce(max($1, $2), 0)'))

    def test_accumulate(self):
        self.assertEqual(
            ['a', 'aa', 'aab', 'aaba', 'aabaa'],
            self.eval('[a,a,b,a,a].accumulate($1 + $2)'))

        self.assertEqual(
            [1],
            self.eval('[].accumulate($1 + $2, 1)'))

    def test_default_if_empty(self):
        self.assertEqual(
            [1, 2],
            self.eval('[].defaultIfEmpty([1, 2])'))

        self.assertEqual(
            [3, 4],
            self.eval('[3, 4].defaultIfEmpty([1, 2])'))

        self.assertEqual(
            [1, 2],
            self.eval('[].select($).defaultIfEmpty([1, 2])'))

        self.assertEqual(
            [3, 4],
            self.eval('[3, 4].select($).defaultIfEmpty([1, 2])'))

    def test_generate(self):
        self.assertEqual(
            [0, 2, 4, 6, 8],
            self.eval('generate(0, $ < 10, $ + 2)'))

        self.assertEqual(
            [0, 4, 16, 36, 64],
            self.eval('generate(0, $ < 10, $ + 2, $ * $)'))

    def test_generate_many(self):
        friends = {
            'John': ['Jim'],
            'Jim': ['Jay', 'Jax'],
            'Jax': ['John', 'Jacob', 'Jonathan'],
            'Jacob': ['Jonathan', 'Jenifer'],
        }
        self.assertEqual(
            ['John', 'Jim', 'Jay', 'Jax', 'Jacob', 'Jonathan', 'Jenifer'],
            self.eval(
                'generateMany(John, $data.get($, []), decycle => true)',
                friends))

        self.assertEqual(
            ['John', 'Jim', 'Jay', 'Jax', 'Jacob', 'Jonathan', 'Jenifer'],
            self.eval(
                'generateMany(John, $data.get($, []), '
                'decycle => true, depthFirst => true)', friends))

        self.assertEqual(
            ['Jay'],
            self.eval('generateMany(Jay, $data.get($, []))', friends))

        self.assertEqual(
            ['JAX', 'JOHN', 'JACOB', 'JONATHAN', 'JIM', 'JENIFER', 'JAY'],
            self.eval(
                'generateMany(Jax, $data.get($, []), $.toUpper(), '
                'decycle => true)', friends))

    def test_max(self):
        self.assertEqual(
            0,
            self.eval('[].max(0)'))

        self.assertRaises(
            TypeError,
            self.eval, '[].max()')

        self.assertEqual(
            234,
            self.eval('[44, 234, 23].max()'))

    def test_min(self):
        self.assertEqual(
            0,
            self.eval('[].min(0)'))

        self.assertRaises(
            TypeError,
            self.eval, '[].min()')

        self.assertEqual(
            23,
            self.eval('[44, 234, 23].min()'))

    def test_reverse(self):
        self.assertEqual(
            [9, 4, 1],
            self.eval('range(1, 4).select($*$).reverse()'))

    def test_merge_with(self):
        dict1 = {'a': 1, 'b': 'x', 'c': [1, 2], 'x': {'a': 1}}
        dict2 = {'d': 5, 'b': 'y', 'c': [2, 3], 'x': {'b': 2}}
        self.assertEqual(
            {'a': 1, 'c': [1, 2, 3], 'b': 'y', 'd': 5, 'x': {'a': 1, 'b': 2}},
            self.eval(
                '$.d1.mergeWith($.d2)',
                data={'d1': dict1, 'd2': dict2}))

        dict1 = {'a': 1, 'b': 2, 'c': [1, 2]}
        dict2 = {'d': 5, 'b': 3, 'c': [2, 3]}
        self.assertEqual(
            {'a': 1, 'c': [1, 2, 2, 3], 'b': 3, 'd': 5},
            self.eval(
                '$.d1.mergeWith($.d2, $1 + $2)',
                data={'d1': dict1, 'd2': dict2}))

        self.assertEqual(
            {'a': 1, 'b': 3, 'c': [2, 3], 'd': 5},
            self.eval(
                '$.d1.mergeWith($.d2, $1 + $2, maxLevels => 1)',
                data={'d1': dict1, 'd2': dict2}))

        self.assertEqual(
            {'a': 1, 'b': 2, 'c': [1, 2, 3], 'd': 5},
            self.eval(
                '$.d1.mergeWith($.d2,, min($1, $2))',
                data={'d1': dict1, 'd2': dict2}))

    def test_is_iterable(self):
        self.assertEqual(
            True,
            self.eval('isIterable([])'))
        self.assertEqual(
            True,
            self.eval('isIterable([1,2])'))
        self.assertEqual(
            True,
            self.eval('isIterable(set(1,2))'))
        self.assertEqual(
            False,
            self.eval('isIterable(1)'))
        self.assertEqual(
            False,
            self.eval('isIterable("foo")'))
        self.assertEqual(
            False,
            self.eval('isIterable({"a" => 1})'))

    def test_infinite_collections(self):
        self.assertRaises(
            exceptions.CollectionTooLargeException,
            self.eval, 'len(list(sequence()))')

        self.assertRaises(
            exceptions.CollectionTooLargeException,
            self.eval, 'list(sequence())')

        self.assertRaises(
            exceptions.CollectionTooLargeException,
            self.eval, 'len(dict(sequence().select([$, $])))')

        self.assertRaises(
            exceptions.CollectionTooLargeException,
            self.eval, 'dict(sequence().select([$, $]))')

        self.assertRaises(
            exceptions.CollectionTooLargeException,
            self.eval, 'sequence()')

        self.assertRaises(
            exceptions.CollectionTooLargeException,
            self.eval, 'set(sequence())')



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import codecs
import re

from yaql.language import exceptions


NEVER_MATCHING_RE = '(?!x)x'
ESCAPE_SEQUENCE_RE = re.compile(r'''
    ( \\U........      # 8-digit hex escapes
    | \\u....          # 4-digit hex escapes
    | \\x..            # 2-digit hex escapes
    | \\[0-7]{1,3}     # Octal escapes
    | \\N\{[^}]+\}     # Unicode characters by name
    | \\[\\'"abfnrtv]  # Single-character escapes
    )''', re.UNICODE | re.VERBOSE)


def decode_escapes(s):
    def decode_match(match):
        return codecs.decode(match.group(0), 'unicode-escape')
    return ESCAPE_SEQUENCE_RE.sub(decode_match, s)


# noinspection PyPep8Naming
class Lexer:
    t_ignore = ' \t\r\n'

    literals = '()],}'
    keywords = {
        'true': 'TRUE',
        'false': 'FALSE',
        'null': 'NULL'
    }

    keyword_to_val = {
        'TRUE': True,
        'FALSE': False,
        'NULL': None
    }

    def __init__(self, yaql_operators):
        self._operators_table = yaql_operators.operators
        self.tokens = [
            'KEYWORD_STRING',
            'QUOTED_STRING',
            'NUMBER',
            'FUNC',
            'DOLLAR',
            'INDEXER',
            'MAPPING',
            'MAP'
        ] + list(self.keywords.values())
        for op_symbol, op_record in self._operators_table.items():
            if op_symbol in ('[]', '{}'):
                continue
            lexem_name = op_record[2]
            setattr(self, 't_' + lexem_name, re.escape(op_symbol))
            self.tokens.append(lexem_name)
        self.t_MAPPING = re.escape(yaql_operators.name_value_op) \
            if yaql_operators.name_value_op else NEVER_MATCHING_RE
        self.t_INDEXER = '\\[' \
            if '[]' in self._operators_table else NEVER_MATCHING_RE
        self.t_MAP = '{' \
            if '{}' in self._operators_table else NEVER_MATCHING_RE

    @staticmethod
    def t_DOLLAR(t):
        """
        \\$\\w*
        """
        return t

    @staticmethod
    def t_NUMBER(t):
        """
        \\b\\d+(\\.?\\d+)?\\b
        """
        if '.' in t.value:
            t.value = float(t.value)
        else:
            t.value = int(t.value)
        return t

    @staticmethod
    def t_FUNC(t):
        """
        \\b[^\\W\\d]\\w*\\(
        """
        val = t.value[:-1]
        t.value = val
        return t

    def t_KEYWORD_STRING(self, t):
        """
        (?!__)\\b[^\\W\\d]\\w*\\b
        """
        if t.value in self._operators_table:
            t.type = self._operators_table[t.value][2]
        else:
            t.type = self.keywords.get(t.value, 'KEYWORD_STRING')
            t.value = self.keyword_to_val.get(t.type, t.value)
        return t

    @staticmethod
    def t_QUOTED_STRING(t):
        """
        '([^'\\\\]|\\\\.)*'
        """
        t.value = decode_escapes(t.value[1:-1])
        return t

    @staticmethod
    def t_DOUBLE_QUOTED_STRING(t):
        """
        "([^"\\\\]|\\\\.)*"
        """
        t.value = decode_escapes(t.value[1:-1])
        t.type = 'QUOTED_STRING'
        return t

    @staticmethod
    def t_QUOTED_VERBATIM_STRING(t):
        """
        `([^`\\\\]|\\\\.)*`
        """
        t.value = t.value[1:-1].replace('\\`', '`')
        t.type = 'QUOTED_STRING'
        return t

    @staticmethod
    def t_error(t):
        raise exceptions.YaqlLexicalException(t.value[0], t.lexpos)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pbr.version

from yaql.language import contexts
from yaql.language import conventions
from yaql.language import factory
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes
from yaql.standard_library import boolean as std_boolean
from yaql.standard_library import branching as std_branching
from yaql.standard_library import collections as std_collections
from yaql.standard_library import common as std_common
from yaql.standard_library import date_time as std_datetime
from yaql.standard_library import math as std_math
from yaql.standard_library import queries as std_queries
from yaql.standard_library import regex as std_regex
from yaql.standard_library import strings as std_strings
from yaql.standard_library import system as std_system
from yaql.standard_library import yaqlized as std_yaqlized

_cached_expressions = {}
_cached_engine = None
_default_context = None


def detect_version():
    version_info = pbr.version.VersionInfo('yaql')
    try:
        return version_info.version_string()
    except AttributeError:
        return 'Undefined (package was not installed)'


__version__ = detect_version()


def _setup_context(data, context, finalizer, convention):
    if context is None:
        context = contexts.Context(
            convention=convention or conventions.CamelCaseConvention())

    if finalizer is None:
        @specs.parameter('iterator', yaqltypes.Iterable())
        @specs.name('#iter')
        def limit(iterator):
            return iterator

        @specs.inject('limiter', yaqltypes.Delegate('#iter'))
        @specs.inject('engine', yaqltypes.Engine())
        @specs.name('#finalize')
        def finalize(obj, limiter, engine):
            if engine.options.get('yaql.convertOutputData', True):
                return utils.convert_output_data(obj, limiter, engine)
            return obj

        context.register_function(limit)
        context.register_function(finalize)
    else:
        context.register_function(finalizer)

    if data is not utils.NO_VALUE:
        context['$'] = utils.convert_input_data(data)
    return context


def create_context(data=utils.NO_VALUE, context=None, system=True,
                   common=True, boolean=True, strings=True,
                   math=True, collections=True, queries=True,
                   regex=True, branching=True,
                   no_sets=False, finalizer=None, delegates=False,
                   convention=None, datetime=True, yaqlized=True,
                   group_by_agg_fallback=True):

    context = _setup_context(data, context, finalizer, convention)
    if system:
        std_system.register_fallbacks(context)
        context = context.create_child_context()
        std_system.register(context, delegates)
    if common:
        std_common.register(context)
    if boolean:
        std_boolean.register(context)
    if strings:
        std_strings.register(context)
    if math:
        std_math.register(context)
    if collections:
        std_collections.register(context, no_sets)
    if queries:
        std_queries.register(context, group_by_agg_fallback)
    if regex:
        std_regex.register(context)
    if branching:
        std_branching.register(context)
    if datetime:
        std_datetime.register(context)
    if yaqlized:
        context = std_yaqlized.register(context)

    return context


YaqlFactory = factory.YaqlFactory


def eval(expression, data=None):
    global _cached_engine, _cached_expressions, _default_context

    if _cached_engine is None:
        _cached_engine = YaqlFactory().create()

    parsed_expression = _cached_expressions.get(expression)
    if parsed_expression is None:
        parsed_expression = _cached_engine(expression)
        _cached_expressions[expression] = parsed_expression

    if _default_context is None:
        _default_context = create_context()

    return parsed_expression.evaluate(
        data=data, context=_default_context.create_child_context())



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
The module describes which operations can be done with strings in YAQL.
"""

import string as string_module

from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes


@specs.parameter('left', yaqltypes.String())
@specs.parameter('right', yaqltypes.String())
@specs.name('#operator_>')
def gt(left, right):
    """:yaql:operator >

    Returns true if the left operand is strictly greater than the right,
    ordering lexicographically, otherwise false.

    :signature: left > right
    :arg left: left operand
    :argType left: string
    :arg right: right operand
    :argType right: string
    :returnType: boolean

    .. code::

        yaql> "abc" > "ab"
        true
        yaql> "abc" > "abb"
        true
        yaql> "abc" > "abc"
        false
    """
    return left > right


@specs.parameter('left', yaqltypes.String())
@specs.parameter('right', yaqltypes.String())
@specs.name('#operator_<')
def lt(left, right):
    """:yaql:operator <

    Returns true if the left operand is strictly less than the right, ordering
    lexicographically, otherwise false.

    :signature: left < right
    :arg left: left operand
    :argType left: string
    :arg right: right operand
    :argType right: string
    :returnType: boolean

    .. code::

        yaql> "ab" < "abc"
        true
        yaql> "abb" < "abc"
        true
        yaql> "abc" < "abc"
        false
    """
    return left < right


@specs.parameter('left', yaqltypes.String())
@specs.parameter('right', yaqltypes.String())
@specs.name('#operator_>=')
def gte(left, right):
    """:yaql:operator >=

    Returns true if the left operand is greater or equal to the right, ordering
    lexicographically, otherwise false.

    :signature: left >= right
    :arg left: left operand
    :argType left: string
    :arg right: right operand
    :argType right: string
    :returnType: boolean

    .. code::

        yaql> "abc" >= "ab"
        true
        yaql> "abc" >= "abc"
        true
    """
    return left >= right


@specs.parameter('left', yaqltypes.String())
@specs.parameter('right', yaqltypes.String())
@specs.name('#operator_<=')
def lte(left, right):
    """:yaql:operator <=

    Returns true if the left operand is less or equal to the right, ordering
    lexicographically, otherwise false.

    :signature: left <= right
    :arg left: left operand
    :argType left: string
    :arg right: right operand
    :argType right: string
    :returnType: boolean

    .. code::

        yaql> "ab" <= "abc"
        true
        yaql> "abc" <= "abc"
        true
    """
    return left <= right


@specs.parameter('args', yaqltypes.String())
def concat(*args):
    """:yaql:concat

    Returns concatenated args.

    :signature: concat([args])
    :arg [args]: values to be joined
    :argType [args]: string
    :returnType: string

    .. code::

        yaql> concat("abc", "de", "f")
        "abcdef"
    """
    return ''.join(args)


@specs.parameter('string', yaqltypes.String())
@specs.method
def to_upper(string):
    """:yaql:toUpper

    Returns a string with all case-based characters uppercase.

    :signature: string.toUpper()
    :receiverArg string: value to uppercase
    :argType string: string
    :returnType: string

    .. code::

        yaql> "aB1c".toUpper()
        "AB1C"
    """
    return string.upper()


@specs.parameter('string', yaqltypes.String())
@specs.extension_method
def len_(string):
    """:yaql:len

    Returns size of the string.

    :signature: string.len()
    :receiverArg string: input string
    :argType string: string
    :returnType: integer

    .. code::

        yaql> "abc".len()
        3
    """
    return len(string)


@specs.parameter('string', yaqltypes.String())
@specs.method
def to_lower(string):
    """:yaql:toLower

    Returns a string with all case-based characters lowercase.

    :signature: string.toLower()
    :receiverArg string: value to lowercase
    :argType string: string
    :returnType: string

    .. code::

        yaql> "AB1c".toLower()
        "ab1c"
    """
    return string.lower()


@specs.parameter('string', yaqltypes.String())
@specs.parameter('separator', yaqltypes.String(nullable=True))
@specs.parameter('max_splits', int)
@specs.method
def split(string, separator=None, max_splits=-1):
    """:yaql:split

    Returns a list of tokens in the string, using separator as the
    delimiter.

    :signature: string.split(separator => null, maxSplits => -1)
    :receiverArg string: value to be splitted
    :argType string: string
    :arg separator: delimiter for splitting. null by default, which means
        splitting with whitespace characters
    :argType separator: string
    :arg maxSplits: maximum number of splittings. -1 by default, which means
        all possible splits are done
    :argType maxSplits: integer
    :returnType: list

    .. code::

        yaql> "abc     de  f".split()
        ["abc", "de", "f"]
        yaql> "abc     de  f".split(maxSplits => 1)
        ["abc", "de  f"]
        yaql> "abcde".split("c")
        ["ab", "de"]
    """
    return string.split(separator, max_splits)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('separator', yaqltypes.String(nullable=True))
@specs.parameter('max_splits', int)
@specs.method
def right_split(string, separator=None, max_splits=-1):
    """:yaql:rightSplit

    Returns a list of tokens in the string, using separator as the
    delimiter. If maxSplits is given then at most maxSplits splits are done -
    the rightmost ones.

    :signature: string.rightSplit(separator => null, maxSplits => -1)
    :receiverArg string: value to be splitted
    :argType string: string
    :arg separator: delimiter for splitting. null by default, which means
        splitting with whitespace characters
    :argType separator: string
    :arg maxSplits: number of splits to be done - the rightmost ones.
        -1 by default, which means all possible splits are done
    :argType maxSplits: integer
    :returnType: list

    .. code::

        yaql> "abc     de  f".rightSplit()
        ["abc", "de", "f"]
        yaql> "abc     de  f".rightSplit(maxSplits => 1)
        ["abc     de", "f"]
    """
    return string.rsplit(separator, max_splits)


@specs.parameter('sequence', yaqltypes.Iterable())
@specs.parameter('separator', yaqltypes.String())
@specs.inject('str_delegate', yaqltypes.Delegate('str'))
@specs.method
def join(sequence, separator, str_delegate):
    """:yaql:join

    Returns a string with sequence elements joined by the separator.

    :signature: sequence.join(separator)
    :receiverArg sequence: chain of values to be joined
    :argType sequence: sequence of strings
    :arg separator: value to be placed between joined pairs
    :argType separator: string
    :returnType: string

    .. code::

        yaql> ["abc", "de", "f"].join("")
        "abcdef"
        yaql> ["abc", "de", "f"].join("|")
        "abc|de|f"
    """
    return separator.join(map(str_delegate, sequence))


@specs.parameter('sequence', yaqltypes.Iterable())
@specs.parameter('separator', yaqltypes.String())
@specs.inject('str_delegate', yaqltypes.Delegate('str'))
@specs.method
def join_(separator, sequence, str_delegate):
    """:yaql:join

    Returns a string with sequence elements joined by the separator.

    :signature: separator.join(sequence)
    :receiverArg separator: value to be placed between joined pairs
    :argType separator: string
    :arg sequence: chain of values to be joined
    :argType sequence: sequence of strings
    :returnType: string

    .. code::

        yaql> "|".join(["abc", "de", "f"])
        "abc|de|f"
    """
    return join(sequence, separator, str_delegate)


@specs.parameter('value', nullable=True)
def str_(value):
    """:yaql:str

    Returns a string representation of the value.

    :signature: str(value)
    :arg value: value to be evaluated to string
    :argType value: any
    :returnType: string

    .. code::

        yaql> str(["abc", "de"])
        "(u'abc', u'd')"
        yaql> str(123)
        "123"
    """
    if value is None:
        return 'null'
    elif value is True:
        return 'true'
    elif value is False:
        return 'false'
    else:
        return str(value)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('chars', yaqltypes.String(nullable=True))
@specs.method
def trim(string, chars=None):
    """:yaql:trim

    Returns a string with the leading and trailing chars removed.

    :signature: string.trim(chars => null)
    :receiverArg string: value to be trimmed
    :argType string: string
    :arg chars: symbols to be removed from input string. null by default,
        which means trim is done with whitespace characters
    :argType chars: string
    :returnType: string

    .. code::

        yaql> "  abcd ".trim()
        "abcd"
        yaql> "aababa".trim("a")
        "bab"
    """
    return string.strip(chars)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('chars', yaqltypes.String(nullable=True))
@specs.method
def trim_left(string, chars=None):
    """:yaql:trimLeft

    Returns a string with the leading chars removed.

    :signature: string.trimLeft(chars => null)
    :receiverArg string: value to be trimmed
    :argType string: string
    :arg chars: symbols to be removed from start of input string. null by
        default, which means trim is done with whitespace characters
    :argType chars: string
    :returnType: string

    .. code::

        yaql> "  abcd ".trimLeft()
        "abcd "
        yaql> "aababa".trimLeft("a")
        "baba"
    """
    return string.lstrip(chars)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('chars', yaqltypes.String(nullable=True))
@specs.method
def trim_right(string, chars=None):
    """:yaql:trimRight

    Returns a string with the trailing chars removed.

    :signature: string.trimRight(chars => null)
    :receiverArg string: value to be trimmed
    :argType string: string
    :arg chars: symbols to be removed from end of input string. null by
        default, which means trim is done with whitespace characters
    :argType chars: string
    :returnType: string

    .. code::

        yaql> "  abcd ".trimRight()
        "  abcd"
        yaql> "aababa".trimRight("a")
        "aabab"
    """
    return string.rstrip(chars)


@specs.parameter('string', yaqltypes.String(nullable=True))
@specs.parameter('chars', yaqltypes.String(nullable=True))
@specs.extension_method
def norm(string, chars=None):
    """:yaql:norm

    Returns a string with the leading and trailing chars removed.
    If the resulting string is empty, returns null.

    :signature: string.norm(chars => null)
    :receiverArg string: value to be cut with specified chars
    :argType string: string
    :arg chars: symbols to be removed from the start and the end of input
        string. null by default, which means norm is done with whitespace
        characters
    :argType chars: string
    :returnType: string

    .. code::

        yaql> "  abcd ".norm()
        "abcd"
        yaql> "aaaa".norm("a")
        null
    """
    if string is None:
        return None
    value = string.strip(chars)
    return None if not value else value


@specs.parameter('string', yaqltypes.String(nullable=True))
@specs.parameter('trim_spaces', bool, alias='trim')
@specs.parameter('chars', yaqltypes.String(nullable=True))
@specs.extension_method
def is_empty(string, trim_spaces=True, chars=None):
    """:yaql:isEmpty

    Returns true if the string with removed leading and trailing chars is
    empty.

    :signature: string.isEmpty(trimSpaces => true, chars => null)
    :receiverArg string: value to be checked for emptiness after trim
    :argType string: string
    :arg trimSpaces: true by default, which means string to be trimmed with
        chars. false means checking whether input string is empty
    :argType trimSpaces: boolean
    :arg chars: symbols for trimming. null by default, which means trim is
        done with whitespace characters
    :argType chars: string
    :returnType: boolean

    .. code::

        yaql> "abaab".isEmpty(chars=>"ab")
        true
        yaql> "aba".isEmpty(chars=>"a")
        false
    """
    if string is None:
        return True
    if trim_spaces:
        string = string.strip(chars)
    return not string


@specs.parameter('string', yaqltypes.String())
@specs.parameter('old', yaqltypes.String())
@specs.parameter('new', yaqltypes.String())
@specs.parameter('count', int)
@specs.method
def replace(string, old, new, count=-1):
    """:yaql:replace

    Returns a string with first count occurrences of old replaced with new.

    :signature: string.replace(old, new, count => -1)
    :receiverArg string: input string
    :argType string: string
    :arg old: value to be replaced
    :argType old: string
    :arg new: replacement for old value
    :argType new: string
    :arg count: how many first replacements to do. -1 by default, which means
        to do all replacements
    :argType count: integer
    :returnType: string

    .. code::

        yaql> "abaab".replace("ab", "cd")
        "cdacd"
    """
    return string.replace(old, new, count)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('replacements', utils.MappingType)
@specs.parameter('count', int)
@specs.inject('str_func', yaqltypes.Delegate('str'))
@specs.method
@specs.name('replace')
def replace_with_dict(string, str_func, replacements, count=-1):
    """:yaql:replace

    Returns a string with all occurrences of replacements' keys replaced
    with corresponding replacements' values.
    If count is specified, only the first count occurrences of every key
    are replaced.

    :signature: string.replace(replacements, count => -1)
    :receiverArg string: input string
    :argType string: string
    :arg replacements: dict of replacements in format {old => new ...}
    :argType replacements: mapping
    :arg count: how many first occurrences of every key are replaced. -1 by
        default, which means to do all replacements
    :argType count: integer
    :returnType: string

    .. code::

        yaql> "abc ab abc".replace({abc => xx, ab => yy})
        "xx yy xx"
        yaql> "abc ab abc".replace({ab => yy, abc => xx})
        "yyc yy yyc"
        yaql> "abc ab abc".replace({ab => yy, abc => xx}, 1)
        "yyc ab xx"
    """
    for key, value in replacements.items():
        string = string.replace(str_func(key), str_func(value), count)
    return string


@specs.parameter('left', yaqltypes.String())
@specs.parameter('right', int)
@specs.name('#operator_*')
def string_by_int(left, right, engine):
    """:yaql:operator *

    Returns string repeated count times.

    :signature: left * right
    :arg left: left operand
    :argType left: string
    :arg right: right operator, how many times repeat input string
    :argType right: integer
    :returnType: string

    .. code::

        yaql> "ab" * 2
        "abab"
    """
    utils.limit_memory_usage(engine, (-right + 1, ''), (right, left))
    return left * right


@specs.parameter('left', yaqltypes.String())
@specs.parameter('right', yaqltypes.String())
@specs.name('#operator_in')
def in_(left, right):
    """:yaql:operator in

    Returns true if there is at least one occurrence of left string in right.

    :signature: left in right
    :arg left: left operand, which occurrence is checked
    :argType left: string
    :arg right: right operand
    :argType right: string
    :returnType: boolean

    .. code::

        yaql> "ab" in "abc"
        true
        yaql> "ab" in "acb"
        false
    """
    return left in right


@specs.parameter('left', int)
@specs.parameter('right', yaqltypes.String())
@specs.name('#operator_*')
def int_by_string(left, right, engine):
    """:yaql:operator *

    Returns string repeated count times.

    :signature: left * right
    :arg left: left operand, how many times repeat input string
    :argType left: integer
    :arg right: right operator
    :argType right: string
    :returnType: string

    .. code::

        yaql> 2 * "ab"
        "abab"
    """
    return string_by_int(right, left, engine)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('start', int)
@specs.parameter('length', int)
@specs.method
def substring(string, start, length=-1):
    """:yaql:substring

    Returns a substring beginning from start index ending with start+end index.

    :signature: string.substring(start, length => -1)
    :receiverArg string: input string
    :argType string: string
    :arg start: index for substring to start with
    :argType start: integer
    :arg length: length of substring. -1 by default, which means end of
        substring to be equal to the end of input string
    :argType length: integer
    :returnType: string

    .. code::

        yaql> "abcd".substring(1)
        "bcd"
        yaql> "abcd".substring(1, 2)
        "bc"
    """
    if length < 0:
        length = len(string)
    if start < 0:
        start += len(string)
    return string[start:start + length]


@specs.parameter('string', yaqltypes.String())
@specs.parameter('sub', yaqltypes.String())
@specs.parameter('start', int)
@specs.method
def index_of(string, sub, start=0):
    """:yaql:indexOf

    Returns an index of first occurrence sub in string beginning from start.
    -1 is a return value if there is no any occurrence.

    :signature: string.indexOf(sub, start => 0)
    :receiverArg string: input string
    :argType string: string
    :arg sub: substring to find in string
    :argType sub: string
    :arg start: index to start search with, 0 by default
    :argType start: integer
    :returnType: integer

    .. code::

        yaql> "cabcdab".indexOf("ab")
        1
        yaql> "cabcdab".indexOf("ab", 2)
        5
        yaql> "cabcdab".indexOf("ab", 6)
        -1
    """
    return string.find(sub, start)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('sub', yaqltypes.String())
@specs.parameter('start', int)
@specs.parameter('length', int)
@specs.method
def index_of_(string, sub, start, length):
    """:yaql:indexOf

    Returns an index of first occurrence sub in string beginning from start
    ending with start+length.
    -1 is a return value if there is no any occurrence.

    :signature: string.indexOf(sub, start, length)
    :receiverArg string: input string
    :argType string: string
    :arg sub: substring to find in string
    :argType sub: string
    :arg start: index to start search with, 0 by default
    :argType start: integer
    :arg length: length of string to find substring in
    :argType length: integer
    :returnType: integer

    .. code::

        yaql> "cabcdab".indexOf("bc", 2, 2)
        2
    """
    if start < 0:
        start += len(string)
    if length < 0:
        length = len(string) - start
    return string.find(sub, start, start + length)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('sub', yaqltypes.String())
@specs.parameter('start', int)
@specs.method
def last_index_of(string, sub, start=0):
    """:yaql:lastIndexOf

    Returns an index of last occurrence sub in string beginning from start.
    -1 is a return value if there is no any occurrence.

    :signature: string.lastIndexOf(sub, start => 0)
    :receiverArg string: input string
    :argType string: string
    :arg sub: substring to find in string
    :argType sub: string
    :arg start: index to start search with, 0 by default
    :argType start: integer
    :returnType: integer

    .. code::

        yaql> "cabcdab".lastIndexOf("ab")
        5
    """
    return string.rfind(sub, start)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('sub', yaqltypes.String())
@specs.parameter('start', int)
@specs.parameter('length', int)
@specs.method
def last_index_of_(string, sub, start, length):
    """:yaql:lastIndexOf

    Returns an index of last occurrence sub in string beginning from start
    ending with start+length.
    -1 is a return value if there is no any occurrence.

    :signature: string.lastIndexOf(sub, start, length)
    :receiverArg string: input string
    :argType string: string
    :arg sub: substring to find in string
    :argType sub: string
    :arg start: index to start search with, 0 by default
    :argType start: integer
    :arg length: length of string to find substring in
    :argType length: integer
    :returnType: integer

    .. code::

        yaql> "cabcdbc".lastIndexOf("bc", 2, 5)
        5
    """
    if start < 0:
        start += len(string)
    if length < 0:
        length = len(string) - start
    return string.rfind(sub, start, start + length)


@specs.parameter('string', yaqltypes.String())
@specs.method
def to_char_array(string):
    """:yaql:toCharArray

    Converts a string to array of one character strings.

    :signature: string.toCharArray()
    :receiverArg string: input string
    :argType string: string
    :returnType: list

    .. code::

        yaql> "abc de".toCharArray()
        ["a", "b", "c", " ", "d", "e"]
    """
    return tuple(string)


def characters(
        digits=False, hexdigits=False,
        ascii_lowercase=False, ascii_uppercase=False,
        ascii_letters=False, letters=False,
        octdigits=False, punctuation=False, printable=False,
        lowercase=False, uppercase=False, whitespace=False):
    """:yaql:characters

    Returns a list of all distinct items of specified types.

    :signature: characters(digits => false, hexdigits => false,
                           asciiLowercase => false, asciiUppercase => false,
                           asciiLetters => false, letters => false,
                           octdigits => false, punctuation => false,
                           printable => false, lowercase => false,
                           uppercase => false, whitespace => false)
    :arg digits: include digits in output list if true, false by default
    :argType digits: boolean
    :arg hexdigits: include hexademical digits in output list if true, false
        by default
    :argType hexdigits: boolean
    :arg asciiLowercase: include ASCII lowercase letters in output list if
        true, false by default
    :argType asciiLowercase: boolean
    :arg asciiUppercase: include ASCII uppercase letters in output list if
        true, false by default
    :argType asciiUppercase: boolean
    :arg asciiLetters: include both ASCII lowercase and uppercase letters
        in output list if true, false by default
    :argType asciiLetters: boolean
    :arg letters: include both lowercase and uppercase letters in output list
        if true, false by default
    :argType letters: boolean
    :arg octdigits: include digits from 0 to 7 in output list if true, false
        by default
    :argType octdigits: boolean
    :arg punctuation: include ASCII characters, which are considered
        punctuation, in output list if true, false by default
    :argType punctuation: boolean
    :arg printable: include digits, letters, punctuation, and whitespace in
        output list if true, false by default
    :argType printable: boolean
    :arg lowercase: include lowercase letters in output list if true, false
        by default
    :argType lowercase: boolean
    :arg uppercase: include uppercase letters in output list if true, false
        by default
    :argType uppercase: boolean
    :arg whitespace: include all characters that are considered whitespace
        in output list if true, false by default
    :argType whitespace: boolean
    :returnType: list

    .. code::

        yaql> characters(digits => true)
        ["1", "0", "3", "2", "5", "4", "7", "6", "9", "8"]

    """
    string = ''
    if digits:
        string += string_module.digits
    if hexdigits:
        string += string_module.hexdigits
    if ascii_lowercase:
        string += string_module.ascii_lowercase
    if ascii_uppercase:
        string += string_module.ascii_uppercase
    if ascii_letters:
        string += string_module.ascii_letters
    if letters:
        string += string_module.letters
    if octdigits:
        string += string_module.octdigits
    if punctuation:
        string += string_module.punctuation
    if printable:
        string += string_module.printable
    if lowercase:
        string += string_module.lowercase
    if uppercase:
        string += string_module.uppercase
    if whitespace:
        string += string_module.whitespace
    return tuple(set(string))


def is_string(arg):
    """:yaql:isString

    Returns true if arg is a string.

    :signature: isString(arg)
    :arg arg: input value
    :argType arg: any
    :returnType: boolean

    .. code::

        yaql> isString("ab")
        true
        yaql> isString(1)
        false
    """
    return isinstance(arg, str)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('prefixes', yaqltypes.String())
@specs.method
def starts_with(string, *prefixes):
    """:yaql:startsWith

    Returns true if a string starts with any of given args.

    :signature: string.startsWith([args])
    :receiverArg string: input string
    :argType string: string
    :arg [args]: chain of strings to check input string with
    :argType [args]: strings
    :returnType: boolean

    .. code::

        yaql> "abcd".startsWith("ab", "xx")
        true
        yaql> "abcd".startsWith("yy", "xx", "zz")
        false
    """
    return string.startswith(prefixes)


@specs.parameter('string', yaqltypes.String())
@specs.parameter('suffixes', yaqltypes.String())
@specs.method
def ends_with(string, *suffixes):
    """:yaql:endsWith

    Returns true if a string ends with any of given args.

    :signature: string.endsWith([args])
    :receiverArg string: input string
    :argType string: string
    :arg [args]: chain of strings to check input string with
    :argType [args]: strings
    :returnType: boolean

    .. code::

        yaql> "abcd".endsWith("cd", "xx")
        true
        yaql> "abcd".endsWith("yy", "xx", "zz")
        false
    """
    return string.endswith(suffixes)


@specs.parameter('num', yaqltypes.Number(nullable=True))
def hex_(num):
    """:yaql:hex

    Returns a string with hexadecimal representation of num.

    :signature: hex(num)
    :arg num: input number to be converted to hexademical
    :argType num: number
    :returnType: string

    .. code::

        yaql> hex(256)
        "0x100"
    """
    return hex(num)


def register(context):
    context.register_function(gt)
    context.register_function(lt)
    context.register_function(gte)
    context.register_function(lte)
    context.register_function(len_)
    context.register_function(to_lower)
    context.register_function(to_upper)
    context.register_function(split)
    context.register_function(right_split)
    context.register_function(join)
    context.register_function(join_)
    context.register_function(str_)
    context.register_function(concat)
    context.register_function(concat, name='#operator_+')
    context.register_function(trim)
    context.register_function(trim_left)
    context.register_function(trim_right)
    context.register_function(replace)
    context.register_function(replace_with_dict)
    context.register_function(is_empty)
    context.register_function(string_by_int)
    context.register_function(int_by_string)
    context.register_function(substring)
    context.register_function(index_of)
    context.register_function(index_of_)
    context.register_function(last_index_of)
    context.register_function(last_index_of_)
    context.register_function(to_char_array)
    context.register_function(characters)
    context.register_function(is_string)
    context.register_function(norm)
    context.register_function(in_)
    context.register_function(starts_with)
    context.register_function(ends_with)
    context.register_function(hex_)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
The module describes branching functions such as switch, case, and others.
"""

from yaql.language import specs
from yaql.language import yaqltypes


@specs.parameter('args', yaqltypes.MappingRule())
@specs.no_kwargs
def switch(*args):
    """:yaql:switch

    Returns the value of the first argument for which the key evaluates to
    true, null if there is no such arg.

    :signature: switch([args])
    :arg [args]: mappings with keys to check for true and appropriate values
    :argType [args]: chain of mapping
    :returnType: any (types of values of args)

    .. code::

        yaql> switch("ab" > "abc" => 1, "ab" >= "abc" => 2, "ab" < "abc" => 3)
        3
    """
    for mapping in args:
        if mapping.source():
            return mapping.destination()


@specs.parameter('args', yaqltypes.Lambda())
def select_case(*args):
    """:yaql:selectCase

    Returns a zero-based index of the first predicate evaluated to true. If
    there is no such predicate, returns the count of arguments. All the
    predicates after the first one which was evaluated to true remain
    unevaluated.

    :signature: selectCase([args])
    :arg [args]: predicates to check for true
    :argType [args]: chain of predicates
    :returnType: integer

    .. code::

        yaql> selectCase("ab" > "abc", "ab" >= "abc", "ab" < "abc")
        2
    """
    index = 0
    for f in args:
        if f():
            return index
        index += 1
    return index


@specs.parameter('args', yaqltypes.Lambda())
def select_all_cases(*args):
    """:yaql:selectAllCases

    Evaluates input predicates and returns an iterator to collection of
    zero-based indexes of predicates which were evaluated to true. The actual
    evaluation is done lazily as the iterator advances, not during the
    function call.

    :signature: selectAllCases([args])
    :arg [args]: predicates to check for true
    :argType [args]: chain of predicates
    :returnType: iterator

    .. code::

        yaql> selectAllCases("ab" > "abc", "ab" <= "abc", "ab" < "abc")
        [1, 2]
    """
    for i, f in enumerate(args):
        if f():
            yield i


@specs.parameter('args', yaqltypes.Lambda())
def examine(*args):
    """:yaql:examine

    Evaluates predicates one by one and casts the evaluation results to
    boolean. Returns an iterator to collection of casted results. The actual
    evaluation is done lazily as the iterator advances, not during the
    function call.

    :signature: examine([args])
    :arg [args]: predicates to be evaluated
    :argType [args]: chain of predicates functions
    :returnType: iterator

    .. code::

        yaql> examine("ab" > "abc", "ab" <= "abc", "ab" < "abc")
        [false, true, true]
    """
    for f in args:
        yield bool(f())


@specs.parameter('case', int)
@specs.parameter('args', yaqltypes.Lambda())
@specs.method
def switch_case(case, *args):
    """:yaql:switchCase

    Returns evaluated `case`-th argument. If case is less than 0 or greater
    than the amount of predicates, returns evaluated last argument.
    Returns null if no args are provided.

    :signature: case.switchCase([args])
    :recieverArg case: index of predicate to be evaluated
    :argType case: integer
    :arg [args]: input predicates
    :argType [args]: chain of any types
    :returnType: any

    .. code::

        yaql> 1.switchCase('a', 1 + 1, [])
        2
        yaql> 2.switchCase('a', 1 + 1, [])
        []
        yaql> 3.switchCase('a', 1 + 1, [])
        []
        yaql> let(1) -> selectCase($ < 0, $ = 0).switchCase("less than 0",
                                                            "equal to 0",
                                                            "greater than 0")
        "greater than 0"
    """
    if 0 <= case < len(args):
        return args[case]()
    if len(args) == 0:
        return None
    return args[-1]()


@specs.parameter('args', yaqltypes.Lambda())
def coalesce(*args):
    """:yaql:coalesce

    Returns the first predicate which evaluates to non-null value. Returns null
    if no arguments are provided or if all of them are null.

    :signature: coalesce([args])
    :arg [args]: input arguments
    :argType [args]: chain of any types
    :returnType: any

    .. code::

        yaql> coalesce(null)
        null
        yaql> coalesce(null, [1, 2, 3][0], "abc")
        1
        yaql> coalesce(null, false, 1)
        false
    """
    for f in args:
        res = f()
        if res is not None:
            return res
    return None


def register(context):
    context.register_function(switch)
    context.register_function(select_case)
    context.register_function(switch_case)
    context.register_function(select_all_cases)
    context.register_function(examine)
    context.register_function(coalesce)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Whenever an expression is used in the context of boolean operations, the
following values are interpreted as false: ``false``, ``null``, numeric zero of
any type, empty strings, empty dict, empty list, empty set, zero timespan.
All other values are interpreted as true.
"""


from yaql.language import specs
from yaql.language import yaqltypes


@specs.parameter('left', yaqltypes.Lambda())
@specs.parameter('right', yaqltypes.Lambda())
@specs.name('#operator_and')
def and_(left, right):
    """:yaql:operator and

    Returns left operand if it evaluates to false. Otherwise evaluates right
    operand and returns it.

    :signature: left and right
    :arg left: left operand
    :argType left: any
    :arg right: right operand
    :argType right: any
    :returnType: any (left or right operand types)

    .. code::

        yaql> 1 and 0
        0
        yaql> 1 and 2
        2
        yaql> [] and 1
        []
    """
    return left() and right()


@specs.parameter('left', yaqltypes.Lambda())
@specs.parameter('right', yaqltypes.Lambda())
@specs.name('#operator_or')
def or_(left, right):
    """:yaql:operator or

    Returns left operand if it evaluates to true. Otherwise evaluates right
    operand and returns it.

    :signature: left or right
    :arg left: left operand
    :argType left: any
    :arg right: right operand
    :argType right: any
    :returnType: any (left or right operand types)

    .. code::

        yaql> 1 or 0
        1
        yaql> 1 or 2
        1
        yaql> [] or 1
        1
    """
    return left() or right()


@specs.name('#unary_operator_not')
def not_(arg):
    """:yaql:operator not

    Returns true if arg evaluates to false. Otherwise returns false.

    :signature: not arg
    :arg arg: value to be converted
    :argType arg: any
    :returnType: boolean

    .. code::

        yaql> not true
        false
        yaql> not {}
        true
        yaql> not [1]
        false
    """
    return not arg


def bool_(value):
    """:yaql:bool

    Returns true or false after value type conversion to boolean.
    Function returns false if value is 0, false, empty list, empty dictionary,
    empty string, empty set, and timespan(). All other values are considered
    to be true.

    :signature: bool(value)
    :arg value: value to be converted
    :argType value: any
    :returnType: boolean

    .. code::

        yaql> bool(1)
        true
        yaql> bool([])
        false
    """
    return bool(value)


def is_boolean(value):
    """:yaql:isBoolean

    Returns true if value is boolean, otherwise false.

    :signature: isBoolean(value)
    :arg value: value to check
    :argType value: any
    :returnType: boolean

    .. code::

        yaql> isBoolean(false)
        true
        yaql> isBoolean(0)
        false
    """
    return isinstance(value, bool)


def register(context):
    context.register_function(and_)
    context.register_function(or_)
    context.register_function(not_)
    context.register_function(bool_)
    context.register_function(is_boolean)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect

from yaql.language import exceptions
from yaql.language import utils
from yaql.language import yaqltypes

NO_DEFAULT = utils.create_marker('<NoValue>')


class ParameterDefinition:
    __slots__ = ('value_type', 'name', 'position', 'default', 'alias')

    def __init__(self, name, value_type=None, position=None, alias=None,
                 default=None):
        self.value_type = value_type
        self.name = name
        self.position = position
        self.default = default
        self.alias = alias

    def __repr__(self):
        return '{} => position={} value_type={} default={}'.format(
            self.name, self.position, self.value_type, self.default)

    def clone(self):
        return ParameterDefinition(self.name, self.value_type,
                                   self.position, self.alias, self.default)


class FunctionDefinition:
    __slots__ = ('is_method', 'is_function', 'name', 'parameters', 'payload',
                 'doc', 'no_kwargs', 'meta')

    def __init__(self, name, payload, parameters=None, doc='', meta=None,
                 is_function=True, is_method=False, no_kwargs=False):
        self.is_method = is_method
        self.is_function = is_function
        self.name = name
        self.parameters = {} if not parameters else parameters
        self.payload = payload
        self.doc = doc
        self.no_kwargs = no_kwargs
        self.meta = meta or {}

    def __call__(self, engine, context, receiver=utils.NO_VALUE):
        def func(*args, **kwargs):
            if receiver is not utils.NO_VALUE:
                args = (receiver,) + args
            return self.get_delegate(receiver, engine, context, args, kwargs)()
        return func

    def clone(self):
        parameters = {key: p.clone() for key, p in self.parameters.items()}

        res = FunctionDefinition(
            self.name, self.payload, parameters, self.doc,
            self.meta, self.is_function, self.is_method, self.no_kwargs)
        return res

    def strip_hidden_parameters(self):
        fd = self.clone()
        keys_to_remove = set()

        for k, v in fd.parameters.items():
            if not isinstance(v.value_type, yaqltypes.HiddenParameterType):
                continue
            keys_to_remove.add(k)
            if v.position is not None:
                for v2 in fd.parameters.values():
                    if v2.position is not None and v2.position > v.position:
                        v2.position -= 1
        for key in keys_to_remove:
            del fd.parameters[key]
        return fd

    def set_parameter(self, name, value_type=None, nullable=None,
                      alias=None, overwrite=False):
        if isinstance(name, ParameterDefinition):
            if name.name in self.parameters and not overwrite:
                raise exceptions.DuplicateParameterDecoratorException(
                    function_name=self.name or self.payload.__name__,
                    param_name=name.name)
            self.parameters[name.name] = name
            return name

        spec = inspect.getfullargspec(self.payload)
        if isinstance(name, int):
            if 0 <= name < len(spec.args):
                name = spec.args[name]
            elif name == len(spec.args) and spec.varargs is not None:
                name = spec.varargs
            else:
                raise IndexError('argument position is out of range')

        arg_name = name
        if name == spec.varkw:
            position = None
            arg_name = '**'
        elif name == spec.varargs:
            position = len(spec.args)
            arg_name = '*'
        elif name in spec.kwonlyargs:
            position = None
        elif name not in spec.args:
            raise exceptions.NoParameterFoundException(
                function_name=self.name or self.payload.__name__,
                param_name=name)
        else:
            position = spec.args.index(name)

        default = NO_DEFAULT
        if spec.defaults is not None and name in spec.args:
            index = spec.args.index(name) - len(spec.args)
            if index >= -len(spec.defaults):
                default = spec.defaults[index]
        elif spec.kwonlydefaults is not None:
            default = spec.kwonlydefaults.get(name, NO_DEFAULT)

        if arg_name in self.parameters and not overwrite:
            raise exceptions.DuplicateParameterDecoratorException(
                function_name=self.name or self.payload.__name__,
                param_name=name)

        yaql_type = value_type
        p_nullable = nullable
        if value_type is None:
            if p_nullable is None:
                p_nullable = True
            base_type = object \
                if default in (None, NO_DEFAULT, utils.NO_VALUE) \
                else type(default)
            yaql_type = yaqltypes.PythonType(base_type, p_nullable)
        elif not isinstance(value_type, yaqltypes.SmartType):
            if p_nullable is None:
                p_nullable = default is None
            yaql_type = yaqltypes.PythonType(value_type, p_nullable)

        pd = ParameterDefinition(
            name, yaql_type, position, alias, default
        )
        self.parameters[arg_name] = pd
        return pd

    def insert_parameter(self, name, value_type=None, nullable=None,
                         alias=None, overwrite=False):
        pd = self.set_parameter(name, value_type, nullable, alias, overwrite)
        for p in self.parameters.values():
            if p is pd:
                continue
            if p.position is not None and p.position >= pd.position:
                p.position += 1

    def map_args(self, args, kwargs, context, engine):
        kwargs = dict(kwargs)
        positional_args = len(args) * [
            self.parameters.get('*', utils.NO_VALUE)]
        max_dst_positional_args = len(args) + len(self.parameters)
        positional_fix_table = max_dst_positional_args * [0]
        keyword_args = {}

        for p in self.parameters.values():
            if p.position is not None and isinstance(
                    p.value_type, yaqltypes.HiddenParameterType):
                for index in range(p.position + 1, len(positional_fix_table)):
                    positional_fix_table[index] += 1

        for key, p in self.parameters.items():
            arg_name = p.alias or p.name
            if p.position is not None and key != '*':
                arg_position = p.position - positional_fix_table[p.position]
                if isinstance(p.value_type, yaqltypes.HiddenParameterType):
                    continue
                elif arg_position < len(args) and args[arg_position] \
                        is not utils.NO_VALUE:
                    if arg_name in kwargs:
                        return None
                    positional_args[arg_position] = p
                elif arg_name in kwargs:
                    keyword_args[arg_name] = p
                    del kwargs[arg_name]
                elif p.default is NO_DEFAULT:
                    return None
                elif arg_position < len(args) and args[arg_position]:
                    positional_args[arg_position] = p

            elif p.position is None and key != '**':
                if isinstance(p.value_type, yaqltypes.HiddenParameterType):
                    continue
                elif arg_name in kwargs:
                    keyword_args[arg_name] = p
                    del kwargs[arg_name]
                elif p.default is NO_DEFAULT:
                    return None

        if len(kwargs) > 0:
            if '**' in self.parameters:
                argdef = self.parameters['**']
                for key in kwargs:
                    keyword_args[key] = argdef
            else:
                return None

        for i in range(len(positional_args)):
            if positional_args[i] is utils.NO_VALUE:
                return None
            value = args[i]
            if value is utils.NO_VALUE:
                value = positional_args[i].default
            if not positional_args[i].value_type.check(value, context, engine):
                return None
        for kwd in kwargs:
            if not keyword_args[kwd].value_type.check(
                    kwargs[kwd], context, engine):
                return None

        return tuple(positional_args), keyword_args

    def get_delegate(self, receiver, engine, context, args, kwargs):
        def checked(val, param):
            if not param.value_type.check(val, context, engine):
                raise exceptions.ArgumentException(param.name)

            def convert_arg_func(context2):
                try:
                    return param.value_type.convert(
                        val, receiver, context2, self, engine)
                except exceptions.ArgumentValueException:
                    raise exceptions.ArgumentException(param.name)
            return convert_arg_func

        kwargs = kwargs.copy()
        kwargs = dict(kwargs)
        positional = 0
        for arg_name, p in self.parameters.items():
            if p.position is not None and arg_name != '*':
                positional += 1

        positional_args = positional * [None]
        positional_fix_table = positional * [0]
        keyword_args = {}

        for p in self.parameters.values():
            if p.position is not None and isinstance(
                    p.value_type, yaqltypes.HiddenParameterType):
                for index in range(p.position + 1, positional):
                    positional_fix_table[index] += 1

        for key, p in self.parameters.items():
            arg_name = p.alias or p.name
            if p.position is not None and key != '*':
                if isinstance(p.value_type, yaqltypes.HiddenParameterType):
                    positional_args[p.position] = checked(None, p)
                    positional -= 1
                elif p.position - positional_fix_table[p.position] < len(
                        args) and args[p.position - positional_fix_table[
                            p.position]] is not utils.NO_VALUE:
                    if arg_name in kwargs:
                        raise exceptions.ArgumentException(p.name)
                    positional_args[p.position] = checked(
                        args[p.position - positional_fix_table[
                            p.position]], p)
                elif arg_name in kwargs:
                    positional_args[p.position] = checked(
                        kwargs.pop(arg_name), p)
                elif p.default is not NO_DEFAULT:
                    positional_args[p.position] = checked(p.default, p)
                else:
                    raise exceptions.ArgumentException(p.name)
            elif p.position is None and key != '**':
                if isinstance(p.value_type, yaqltypes.HiddenParameterType):
                    keyword_args[key] = checked(None, p)
                elif arg_name in kwargs:
                    keyword_args[key] = checked(kwargs.pop(arg_name), p)
                elif p.default is not NO_DEFAULT:
                    keyword_args[key] = checked(p.default, p)
                else:
                    raise exceptions.ArgumentException(p.name)
        if len(args) > positional:
            if '*' in self.parameters:
                argdef = self.parameters['*']
                positional_args.extend(
                    map(lambda t: checked(t, argdef), args[positional:]))
            else:
                raise exceptions.ArgumentException('*')
        if len(kwargs) > 0:
            if '**' in self.parameters:
                argdef = self.parameters['**']
                for key, value in kwargs.items():
                    keyword_args[key] = checked(value, argdef)
            else:
                raise exceptions.ArgumentException('**')

        def func():
            new_context = context.create_child_context()
            result = self.payload(
                *tuple(map(lambda t: t(new_context),
                           positional_args)),
                **dict(map(lambda t: (t[0], t[1](new_context)),
                           keyword_args.items()))
            )
            return result

        return func

    def is_valid_method(self):
        min_position = len(self.parameters)
        min_arg = None
        for p in self.parameters.values():
            if p.position is not None and p.position < min_position and \
                    not isinstance(p.value_type,
                                   yaqltypes.HiddenParameterType):
                min_position = p.position
                min_arg = p
        return min_arg and not isinstance(
            min_arg.value_type, yaqltypes.LazyParameterType)


def _get_function_definition(func):
    if not hasattr(func, '__yaql_function__'):
        fd = FunctionDefinition(None, func, {}, func.__doc__)
        func.__yaql_function__ = fd
    return func.__yaql_function__


def _infer_parameter_type(name):
    if name == 'context' or name == '__context':
        return yaqltypes.Context()
    elif name == 'engine' or name == '__engine':
        return yaqltypes.Engine()
    elif name == 'yaql_interface' or name == '__yaql_interface':
        return yaqltypes.YaqlInterface()


def convert_function_name(function_name, convention):
    if not function_name:
        return function_name
    function_name = function_name.rstrip('_')
    if not convention:
        return function_name
    if not function_name[0].isalpha():
        finish = function_name.find(function_name[0], 1)
        if finish <= 1:
            return function_name
        return function_name[:finish + 1] + convention.convert_function_name(
            function_name[finish + 1:])
    return convention.convert_function_name(function_name)


def convert_parameter_name(parameter_name, convention):
    if not parameter_name:
        return parameter_name
    parameter_name = parameter_name.rstrip('_')
    if not convention:
        return parameter_name
    return convention.convert_parameter_name(parameter_name)


def get_function_definition(func, name=None, function=None, method=None,
                            convention=None, parameter_type_func=None):
    if parameter_type_func is None:
        parameter_type_func = _infer_parameter_type
    fd = _get_function_definition(func).clone()
    spec = inspect.getfullargspec(func)
    for arg in spec.args + spec.kwonlyargs:
        if arg not in fd.parameters:
            fd.set_parameter(arg, parameter_type_func(arg))
    if spec.varargs and '*' not in fd.parameters:
        fd.set_parameter(spec.varargs, parameter_type_func(spec.varargs))
    if spec.varkw and '**' not in fd.parameters:
        fd.set_parameter(spec.varkw, parameter_type_func(spec.varkw))

    if name is not None:
        fd.name = name
    elif fd.name is None:
        fd.name = convert_function_name(fd.payload.__name__, convention)
    elif convention is not None:
        fd.name = convert_function_name(fd.name, convention)

    if function is not None:
        fd.is_function = function
    if method is not None:
        fd.is_method = method
    if convention:
        for p in fd.parameters.values():
            if p.alias is None:
                p.alias = convert_parameter_name(p.name, convention)
    return fd


def _parameter(name, value_type=None, nullable=None, alias=None):
    def wrapper(func):
        fd = _get_function_definition(func)
        fd.set_parameter(name, value_type, nullable, alias)
        return func
    return wrapper


def parameter(name, value_type=None, nullable=None, alias=None):
    if value_type is not None and isinstance(
            value_type, yaqltypes.HiddenParameterType):
        raise ValueError('Use inject() for hidden parameters')
    return _parameter(name, value_type, nullable=nullable, alias=alias)


def inject(name, value_type=None, nullable=None, alias=None):
    if value_type is not None and not isinstance(
            value_type, yaqltypes.HiddenParameterType):
        raise ValueError('Use parameter() for normal function parameters')
    return _parameter(name, value_type, nullable=nullable, alias=alias)


def name(function_name):
    def wrapper(func):
        fd = _get_function_definition(func)
        fd.name = function_name
        return func
    return wrapper


def method(func):
    fd = _get_function_definition(func)
    fd.is_method = True
    fd.is_function = False
    return func


def extension_method(func):
    fd = _get_function_definition(func)
    fd.is_method = True
    fd.is_function = True
    return func


def no_kwargs(func):
    fd = _get_function_definition(func)
    fd.no_kwargs = True
    return func


def meta(name, value):
    def wrapper(func):
        fd = _get_function_definition(func)
        fd.meta[name] = value
        return func
    return wrapper


def yaql_property(source_type):
    def decorator(func):
        @name('#property#{}'.format(get_function_definition(func).name))
        @parameter('obj', source_type)
        def wrapper(obj):
            return func(obj)
        return wrapper
    return decorator



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql.tests


class TestCommon(yaql.tests.TestCase):
    def test_null(self):
        self.assertIsNone(self.eval('null'))

    def test_true(self):
        res = self.eval('true')
        self.assertTrue(res)
        self.assertIsInstance(res, bool)

    def test_false(self):
        res = self.eval('false')
        self.assertFalse(res)
        self.assertIsInstance(res, bool)

    def test_string(self):
        self.assertEqual('True', self.eval('True'))
        self.assertEqual('some string', self.eval("'some string'"))

    def test_null_to_null(self):
        self.assertTrue(self.eval('null = null'))
        self.assertFalse(self.eval('null != null'))
        self.assertTrue(self.eval('null <= null'))
        self.assertTrue(self.eval('null >= null'))
        self.assertFalse(self.eval('null < null'))
        self.assertFalse(self.eval('null > null'))

    def test_ordering(self):
        self.assertTrue(self.eval('null < 0'))
        self.assertTrue(self.eval('null < true'))
        self.assertTrue(self.eval('null < false'))
        self.assertTrue(self.eval('null < a'))
        self.assertTrue(self.eval('null <= 0'))
        self.assertFalse(self.eval('null > 0'))
        self.assertFalse(self.eval('null >= 0'))
        self.assertTrue(self.eval('null != 0'))
        self.assertTrue(self.eval('null != false'))
        self.assertFalse(self.eval('null = false'))
        self.assertFalse(self.eval('null = 0'))
        self.assertFalse(self.eval('0 < null'))
        self.assertFalse(self.eval('0 <= null'))
        self.assertTrue(self.eval('0 >= null'))
        self.assertTrue(self.eval('0 > null'))

    def test_max(self):
        self.assertEqual(5, self.eval('max(1, 5)'))
        self.assertEqual(-1, self.eval('max(null, -1)'))
        self.assertIsNone(self.eval('max(null, null)'))

    def test_min(self):
        self.assertEqual(1, self.eval('min(1, 5)'))
        self.assertIsNone(self.eval('min(null, -1)'))
        self.assertIsNone(self.eval('min(null, null)'))

    def test_comparision_of_incomparable(self):
        self.assertFalse(self.eval('a = 1'))
        self.assertFalse(self.eval('a = false'))
        self.assertFalse(self.eval('a = null'))
        self.assertFalse(self.eval('[a] = [false]'))
        self.assertTrue(self.eval('a != 1'))
        self.assertTrue(self.eval('a != false'))
        self.assertTrue(self.eval('[a] != [false]'))
        self.assertTrue(self.eval('a != null'))



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import exceptions
import yaql.tests


class TestLegacyNewEngine(yaql.tests.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval = self.legacy_eval_new_engine

    def test_dict(self):
        self.assertEqual(
            {'a': 'b', 1: 2, None: None},
            self.eval('dict(1 => 2, a => b, null => null)'))

        self.assertEqual({}, self.eval('dict()'))

    def test_list(self):
        self.assertEqual([1, 2, 'a', None], self.eval('list(1, 2, a, null)'))
        self.assertEqual([], self.eval('list()'))
        self.assertEqual([1, 2], self.eval('[1, 2].select($).list()'))

    def test_dict_get(self):
        self.assertEqual(5, self.eval("get($, 'a b')", data={'a b': 5}))
        self.assertEqual(5, self.eval("$.get('a b')", data={'a b': 5}))

    def test_int(self):
        self.assertEqual(5, self.eval("'5'.int()"))
        self.assertEqual(5, self.eval('5.2.int()'))
        self.assertEqual(0, self.eval('null.int()'))

    def test_float(self):
        self.assertAlmostEqual(5.1, self.eval("'5.1'.float()"))
        self.assertAlmostEqual(5.0, self.eval('5.float()'))
        self.assertAlmostEqual(5.1, self.eval("float('5.1')"))
        self.assertAlmostEqual(5.0, self.eval("float(5)"))
        self.assertEqual(0, self.eval('null.float()'))

    def test_bool(self):
        self.assertFalse(self.eval('null.bool()'))
        self.assertFalse(self.eval("''.bool()"))
        self.assertFalse(self.eval('0.bool()'))
        self.assertFalse(self.eval('false.bool()'))
        self.assertFalse(self.eval('[].bool()'))
        self.assertFalse(self.eval('{}.bool()'))
        self.assertTrue(self.eval("' '.bool()"))
        self.assertTrue(self.eval('x.bool()'))
        self.assertTrue(self.eval('1.bool()'))
        self.assertTrue(self.eval('true.bool()'))
        self.assertTrue(self.eval('[1].bool()'))
        self.assertTrue(self.eval('{a=>b}.bool()'))

    def test_filter(self):
        self.assertEqual(2, self.eval("list(1,2,3)[1]"))
        self.assertEqual(3, self.eval("list(1,2,3)[$]", data=2))
        self.assertEqual([1, 3], self.eval("list(1,2,3)[$ != 2]"))
        self.assertEqual([], self.eval("list()[$ > 0]"))

    def test_sum(self):
        self.assertEqual(6, self.eval('list(1,2,3).sum()'))
        self.assertEqual(6, self.eval('sum(list(1,2,3))'))

    def test_range(self):
        self.assertEqual([2, 3, 4, 5], self.eval('range(2).take(4)'))
        self.assertEqual([1, 2, 3], self.eval('range(1, 4)'))
        self.assertEqual([2, 3, 4, 5], self.eval('2.range().take(4)'))
        self.assertEqual([1, 2, 3], self.eval('1.range(4)'))

    def test_take_while(self):
        self.assertEqual([1, 2], self.eval('[1, 2, 3, 4].takeWhile($ < 3)'))
        self.assertEqual([1, 2], self.eval('takeWhile([1, 2, 3, 4], $ < 3)'))

    def test_switch(self):
        expr = 'switch($, $ > 10 => 1, $ <= 10 => -1)'
        self.assertEqual(1, self.eval(expr, data=15))
        self.assertEqual(-1, self.eval(expr, data=5))

    def test_as(self):
        self.assertEqual(
            [3, 6],
            self.eval('[1, 2].as(sum($) => a).select($ * $a)'))

    def test_distinct(self):
        data = [1, 2, 3, 2, 4, 8]
        self.assertEqual([1, 2, 3, 4, 8], self.eval('$.distinct()', data=data))
        self.assertEqual([1, 2, 3, 4, 8], self.eval('distinct($)', data=data))

        data = [{'a': 1}, {'b': 2}, {'a': 1}]
        self.assertEqual(
            [{'a': 1}, {'b': 2}],
            self.eval('$.distinct()', data=data))

    def test_keyword_dict_access(self):
        data = {'A': 12, 'b c': 44, '__d': 99, '_e': 999}
        self.assertEqual(12, self.eval('$.A', data=data))
        self.assertEqual(999, self.eval('$._e', data=data))

        self.assertRaises(exceptions.NoMatchingFunctionException,
                          self.eval, "$.'b c'", data=data)
        self.assertRaises(exceptions.NoMatchingFunctionException,
                          self.eval, '$.123', data=data)
        self.assertIsNone(self.eval('$.b', data=data))
        self.assertRaises(exceptions.YaqlLexicalException,
                          self.eval, '$.__d', data=data)

    def test_compare_not_comparable(self):
        self.assertTrue(self.eval('asd != true'))
        self.assertFalse(self.eval('asd = 0'))


class TestLegacy(TestLegacyNewEngine):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval = self.legacy_eval

    def test_tuples_func(self):
        self.assertEqual((1, 2), self.eval('tuple(1, 2)'))
        self.assertEqual((None,), self.eval('tuple(null)'))
        self.assertEqual((), self.eval('tuple()'))

    def test_tuples(self):
        self.assertEqual((1, 2), self.eval('1 => 2'))
        self.assertEqual((None, 'a b'), self.eval('null => "a b"'))
        self.assertEqual((1, 2, 3), self.eval('1 => 2 => 3'))
        self.assertEqual(((1, 2), 3), self.eval('(1 => 2) => 3'))
        self.assertEqual((1, (2, 3)), self.eval('1 => (2 => 3)'))

    def test_dicts_are_iterable(self):
        data = {'a': 1, 'b': 2}
        self.assertTrue(self.eval('a in $', data))
        self.assertCountEqual('ab', self.eval('$.sum()', data))



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import datetime

from dateutil import tz

from yaql.language import exceptions
from yaql.language import expressions
from yaql.language import utils
from yaql import yaql_interface


class HiddenParameterType(metaclass=abc.ABCMeta):
    __slots__ = tuple()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def check(self, value, context, engine, *args, **kwargs):
        return True


class LazyParameterType(metaclass=abc.ABCMeta):
    __slots__ = tuple()


class SmartType(metaclass=abc.ABCMeta):
    __slots__ = ('nullable',)

    def __init__(self, nullable):
        self.nullable = nullable

    def check(self, value, context, engine, *args, **kwargs):
        if value is None and not self.nullable:
            return False
        return True

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        if not self.check(value, context, engine, *args, **kwargs):
            raise exceptions.ArgumentValueException()
        utils.limit_memory_usage(engine, (1, value))
        return value

    def is_specialization_of(self, other):
        return False


class GenericType(SmartType):
    __slots__ = ('checker', 'converter')

    def __init__(self, nullable, checker=None, converter=None):
        super().__init__(nullable)
        self.checker = checker
        self.converter = converter

    def check(self, value, context, engine, *args, **kwargs):
        if isinstance(value, expressions.Constant):
            value = value.value

        if not super().check(
                value, context, engine, *args, **kwargs):
            return False
        if value is None or isinstance(value, expressions.Expression):
            return True
        if not self.checker:
            return True
        return self.checker(value, context, *args, **kwargs)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        if isinstance(value, expressions.Constant):
            value = value.value
        super().convert(
            value, receiver, context, function_spec, engine, *args, **kwargs)
        if value is None or not self.converter:
            return value
        return self.converter(value, receiver, context, function_spec, engine,
                              *args, **kwargs)


class PythonType(GenericType):
    __slots__ = ('python_type', 'validators')

    def __init__(self, python_type, nullable=True, validators=None):
        self.python_type = python_type
        if not validators:
            validators = [lambda _: True]
        if not isinstance(validators, (list, tuple)):
            validators = [validators]
        self.validators = validators

        super().__init__(
            nullable,
            lambda value, context, *args, **kwargs: isinstance(
                value, self.python_type) and all(
                map(lambda t: t(value), self.validators)))

    def is_specialization_of(self, other):
        if not isinstance(other, PythonType):
            return False
        try:
            len(self.python_type)
            len(other.python_type)
        except Exception:
            return (
                issubclass(self.python_type, other.python_type)
                and not issubclass(other.python_type, self.python_type)
            )
        else:
            return False


class MappingRule(LazyParameterType, SmartType):
    __slots__ = tuple()

    def __init__(self):
        super().__init__(False)

    def check(self, value, context, *args, **kwargs):
        return isinstance(value, expressions.MappingRuleExpression)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        super().convert(
            value, receiver, context, function_spec, engine, *args, **kwargs)
        wrap = lambda func: lambda: func(receiver, context, engine)  # noqa

        return utils.MappingRule(wrap(value.source), wrap(value.destination))


class String(PythonType):
    __slots__ = tuple()

    def __init__(self, nullable=False):
        super().__init__(str, nullable=nullable)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        value = super().convert(
            value, receiver, context, function_spec, engine, *args, **kwargs)
        return None if value is None else str(value)


class Integer(PythonType):
    __slots__ = tuple()

    def __init__(self, nullable=False):
        super().__init__(
            int, nullable=nullable,
            validators=[lambda t: not isinstance(t, bool)])


class DateTime(PythonType):
    __slots__ = tuple()

    utctz = tz.tzutc()

    def __init__(self, nullable=False):
        super().__init__(datetime.datetime, nullable=nullable)

    def convert(self, value, *args, **kwargs):
        if isinstance(value, datetime.datetime):
            if value.tzinfo is None:
                return value.replace(tzinfo=self.utctz)
            else:
                return value
        return super().convert(value, *args, **kwargs)


class Iterable(PythonType):
    __slots__ = tuple()

    def __init__(self, validators=None, nullable=False):
        super().__init__(
            collections.abc.Iterable, nullable,
            [lambda t: not isinstance(t, (str, utils.MappingType))] + (
                validators or []))

    def check(self, value, context, engine, *args, **kwargs):
        if isinstance(value, utils.MappingType) and engine.options.get(
                'yaql.iterableDicts', False):
            return True
        return super().check(
            value, context, engine, *args, **kwargs)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        res = super().convert(
            value, receiver, context, function_spec, engine, *args, **kwargs)
        return None if res is None else utils.limit_iterable(res, engine)


class Iterator(Iterable):
    __slots__ = tuple()

    def __init__(self, validators=None, nullable=False):
        super().__init__(
            validators=[utils.is_iterator] + (validators or []),
            nullable=nullable)


class Sequence(PythonType):
    __slots__ = tuple()

    def __init__(self, validators=None, nullable=False):
        super().__init__(
            collections.abc.Sequence, nullable, [
                lambda t: not isinstance(t, (str, dict))] + (
                    validators or []))


class Number(PythonType):
    __slots__ = tuple()

    def __init__(self, nullable=False):
        super().__init__(
            (int, float), nullable,
            validators=[lambda t: not isinstance(t, bool)])


class Lambda(LazyParameterType, SmartType):
    __slots__ = ('with_context', 'method')

    def __init__(self, with_context=False, method=False):
        super().__init__(True)
        self.with_context = with_context
        self.method = method

    def check(self, value, context, *args, **kwargs):
        if self.method and isinstance(
                value, expressions.Expression) and not value.uses_receiver:
            return False
        return super().check(value, context, *args, **kwargs)

    @staticmethod
    def _publish_params(context, args, kwargs):
        for i, param in enumerate(args):
            context['$' + str(i + 1)] = param
        for arg_name, arg_value in kwargs.items():
            context['$' + arg_name] = arg_value

    def _call(self, value, receiver, context, engine, args, kwargs):
        self._publish_params(context, args, kwargs)
        if isinstance(value, expressions.Expression):
            result = value(receiver, context, engine)
        elif callable(value):
            result = value(*args, **kwargs)
        else:
            result = value
        return result

    def convert(self, value, receiver, context, function_spec, engine,
                *convert_args, **convert_kwargs):
        super().convert(
            value, receiver, context, function_spec, engine,
            *convert_args, **convert_kwargs)
        if value is None:
            return None
        elif callable(value) and hasattr(value, '__unwrapped__'):
            value = value.__unwrapped__

        def func(*args, **kwargs):
            if self.method and self.with_context:
                new_receiver, new_context = args[:2]
                args = args[2:]
            elif self.method and not self.with_context:
                new_receiver, new_context = \
                    args[0], context.create_child_context()
                args = args[1:]
            elif not self.method and self.with_context:
                new_receiver, new_context = utils.NO_VALUE, args[0]
                args = args[1:]
            else:
                new_receiver, new_context = \
                    utils.NO_VALUE, context.create_child_context()

            return self._call(value, new_receiver, new_context,
                              engine, args, kwargs)

        func.__unwrapped__ = value
        return func


class Super(HiddenParameterType, SmartType):
    __slots__ = ('with_context', 'method', 'with_name')

    def __init__(self, with_context=False, method=None, with_name=False):
        self.with_context = with_context
        self.method = method
        self.with_name = with_name
        super().__init__(False)

    @staticmethod
    def _find_function_context(spec, context):
        while context is not None:
            if spec in context:
                return context
            context = context.parent
        raise exceptions.NoFunctionRegisteredException(
            spec.name)

    def convert(self, value, receiver, context, function_spec, engine,
                *convert_args, **convert_kwargs):
        if callable(value) and hasattr(value, '__unwrapped__'):
            value = value.__unwrapped__

        def func(*args, **kwargs):
            function_context = self._find_function_context(
                function_spec, context)
            parent_function_context = function_context.parent
            if parent_function_context is None:
                raise exceptions.NoFunctionRegisteredException(
                    function_spec.name)

            new_name = function_spec.name
            if self.with_name:
                new_name = args[0]
                args = args[1:]

            new_receiver = receiver
            if self.method is True:
                new_receiver = args[0]
                args = args[1:]
            elif self.method is False:
                new_receiver = utils.NO_VALUE

            if self.with_context:
                new_context = args[0]
                args = args[1:]
            else:
                new_context = context.create_child_context()

            return parent_function_context(
                new_name, engine, new_receiver, new_context)(*args, **kwargs)
        func.__unwrapped__ = value
        return func


class Context(HiddenParameterType, SmartType):
    __slots__ = tuple()

    def __init__(self):
        super().__init__(False)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        return context


class Delegate(HiddenParameterType, SmartType):
    __slots__ = ('name', 'with_context', 'method', 'use_convention')

    def __init__(self, name=None, with_context=False, method=False,
                 use_convention=True):
        super().__init__(False)
        self.name = name
        self.with_context = with_context
        self.method = method
        self.use_convention = use_convention

    def convert(self, value, receiver, context, function_spec, engine,
                *convert_args, **convert_kwargs):
        if callable(value) and hasattr(value, '__unwrapped__'):
            value = value.__unwrapped__

        def func(*args, **kwargs):
            name = self.name
            if not name:
                name = args[0]
                args = args[1:]

            new_receiver = utils.NO_VALUE
            if self.method:
                new_receiver = args[0]
                args = args[1:]
            if self.with_context:
                new_context = args[0]
                args = args[1:]
            else:
                new_context = context.create_child_context()

            return new_context(
                name, engine, new_receiver,
                use_convention=self.use_convention)(*args, **kwargs)
        func.__unwrapped__ = value
        return func


class Receiver(HiddenParameterType, SmartType):
    __slots__ = tuple()

    def __init__(self):
        super().__init__(False)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        return receiver


class Engine(HiddenParameterType, SmartType):
    __slots__ = tuple()

    def __init__(self):
        super().__init__(False)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        return engine


class FunctionDefinition(HiddenParameterType, SmartType):
    __slots__ = tuple()

    def __init__(self):
        super().__init__(False)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        return function_spec


class Constant(SmartType):
    __slots__ = ('expand',)

    def __init__(self, nullable, expand=True):
        self.expand = expand
        super().__init__(nullable)

    def check(self, value, context, *args, **kwargs):
        return super().check(
            value, context, *args, **kwargs) and (
            value is None or isinstance(value, expressions.Constant))

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        super().convert(
            value, receiver, context, function_spec, engine, *args, **kwargs)
        return value.value if self.expand else value


class YaqlExpression(LazyParameterType, SmartType):
    __slots__ = ('_expression_types',)

    def __init__(self, expression_type=None):
        super().__init__(False)
        if expression_type and not utils.is_sequence(expression_type):
            expression_type = (expression_type,)
        self._expression_types = expression_type

    def check(self, value, context, *args, **kwargs):
        if not self._expression_types:
            return isinstance(value, expressions.Expression)
        return any(t == type(value) for t in self._expression_types)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        super().convert(
            value, receiver, context, function_spec, engine, *args, **kwargs)
        return value


class StringConstant(Constant):
    __slots__ = tuple()

    def __init__(self, nullable=False):
        super().__init__(nullable)

    def check(self, value, context, *args, **kwargs):
        return super().check(
            value, context, *args, **kwargs) and (
            value is None or isinstance(value.value, str))


class Keyword(Constant):
    __slots__ = tuple()

    def __init__(self, expand=True):
        super().__init__(False, expand)

    def check(self, value, context, *args, **kwargs):
        return isinstance(value, expressions.KeywordConstant)


class BooleanConstant(Constant):
    __slots__ = tuple()

    def __init__(self, nullable=False, expand=True):
        super().__init__(nullable, expand)

    def check(self, value, context, *args, **kwargs):
        return super().check(
            value, context, *args, **kwargs) and (
            value is None or type(value.value) is bool)


class NumericConstant(Constant):
    __slots__ = tuple()

    def __init__(self, nullable=False, expand=True):
        super().__init__(nullable, expand)

    def check(self, value, context, *args, **kwargs):
        return super().check(
            value, context, *args, **kwargs) and (
            value is None or
            isinstance(value.value, (int, float)) and
            type(value.value) is not bool)


class SmartTypeAggregation(SmartType, metaclass=abc.ABCMeta):
    __slots__ = ('types',)

    def __init__(self, *args, **kwargs):
        self.nullable = kwargs.pop('nullable', False)
        super().__init__(self.nullable)

        self.types = []
        for item in args:
            if isinstance(item, (type, tuple)):
                item = PythonType(item)
            if isinstance(item, (HiddenParameterType, LazyParameterType)):
                raise ValueError('Special smart types are not supported')
            self.types.append(item)


class AnyOf(SmartTypeAggregation):
    __slots__ = tuple()

    def _check_match(self, value, context, engine, *args, **kwargs):
        for type_to_check in self.types:
            check_result = type_to_check.check(
                value, context, engine, *args, **kwargs)
            if check_result:
                return type_to_check

    def check(self, value, context, engine, *args, **kwargs):
        if isinstance(value, expressions.Constant):
            value = value.value

        if value is None:
            return self.nullable

        check_result = self._check_match(
            value, context, engine, *args, **kwargs)
        return True if check_result else False

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        if isinstance(value, expressions.Constant):
            value = value.value

        if value is None:
            if self.nullable:
                return None
            else:
                suitable_type = None
        else:
            suitable_type = self._check_match(
                value, context, engine, *args, **kwargs)
        if suitable_type:
            return suitable_type.convert(
                value, receiver, context, function_spec,
                engine, *args, **kwargs)
        raise exceptions.ArgumentValueException()


class Chain(SmartTypeAggregation):
    __slots__ = tuple()

    def _check_match(self, value, context, engine, *args, **kwargs):
        for type_to_check in self.types:
            check_result = type_to_check.check(
                value, context, engine, *args, **kwargs)
            if check_result:
                return type_to_check

    def check(self, value, context, engine, *args, **kwargs):
        if isinstance(value, expressions.Constant):
            value = value.value

        if value is None:
            return self.nullable

        for type_to_check in self.types:
            if not type_to_check.check(
                    value, context, engine, *args, **kwargs):
                return False

        return True

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        if isinstance(value, expressions.Constant):
            value = value.value

        if value is None:
            if self.nullable:
                return None
            raise exceptions.ArgumentValueException()

        for smart_type in self.types:
            value = smart_type.convert(
                value, receiver, context, function_spec,
                engine, *args, **kwargs)
        return value


class NotOfType(SmartType):
    __slots__ = ('smart_type',)

    def __init__(self, smart_type, nullable=True):
        if isinstance(smart_type, (type, tuple)):
            smart_type = PythonType(smart_type, nullable=nullable)
        self.smart_type = smart_type
        super().__init__(nullable)

    def check(self, value, context, engine, *args, **kwargs):
        if isinstance(value, expressions.Constant):
            value = value.value
        if not super().check(
                value, context, engine, *args, **kwargs):
            return False
        if value is None or isinstance(value, expressions.Expression):
            return True
        return not self.smart_type.check(
            value, context, engine, *args, **kwargs)


class YaqlInterface(HiddenParameterType, SmartType):
    __slots__ = tuple()

    def __init__(self):
        super().__init__(False)

    def convert(self, value, receiver, context, function_spec, engine,
                *args, **kwargs):
        return yaql_interface.YaqlInterface(context, engine, receiver)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import re
import uuid

from ply import lex
from ply import yacc

from yaql.language import exceptions
from yaql.language import expressions
from yaql.language import lexer
from yaql.language import parser
from yaql.language import utils


OperatorType = collections.namedtuple('OperatorType', [
    'PREFIX_UNARY', 'SUFFIX_UNARY',
    'BINARY_LEFT_ASSOCIATIVE', 'BINARY_RIGHT_ASSOCIATIVE',
    'NAME_VALUE_PAIR'
])(
    PREFIX_UNARY='PREFIX_UNARY',
    SUFFIX_UNARY='SUFFIX_UNARY',
    BINARY_LEFT_ASSOCIATIVE='BINARY_LEFT_ASSOCIATIVE',
    BINARY_RIGHT_ASSOCIATIVE='BINARY_RIGHT_ASSOCIATIVE',
    NAME_VALUE_PAIR='NAME_VALUE_PAIR'
)


class YaqlOperators:
    def __init__(self, operators, name_value_op=None):
        self.operators = operators
        self.name_value_op = name_value_op


class YaqlEngine:
    def __init__(self, ply_lexer, ply_parser, options, factory):
        self._lexer = ply_lexer
        self._parser = ply_parser
        self._options = utils.FrozenDict(options or {})
        self._factory = factory

    @property
    def lexer(self):
        return self._lexer

    @property
    def parser(self):
        return self._parser

    @property
    def options(self):
        return self._options

    @property
    def factory(self):
        return self._factory

    def __call__(self, expression, options=None):
        if options:
            return self.copy(options)(expression)

        return expressions.Statement(
            self.parser.parse(expression, lexer=self.lexer), self)

    def copy(self, options):
        opt = dict(self._options)
        opt.update(options)
        return YaqlEngine(self._lexer, self._parser, opt, self._factory)


class YaqlFactory:
    def __init__(self, keyword_operator='=>', allow_delegates=False):
        self._keyword_operator = keyword_operator
        self._allow_delegates = allow_delegates
        self.operators = self._standard_operators()
        if keyword_operator:
            self.operators.insert(0, (keyword_operator,
                                      OperatorType.NAME_VALUE_PAIR))

    @property
    def keyword_operator(self):
        return self._keyword_operator

    @property
    def allow_delegates(self):
        return self._allow_delegates

    # noinspection PyMethodMayBeStatic
    def _standard_operators(self):
        return [
            ('.', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('?.', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('[]', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('{}', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('+', OperatorType.PREFIX_UNARY),
            ('-', OperatorType.PREFIX_UNARY),
            (),
            ('=~', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('!~', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('*', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('/', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('mod', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('+', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('-', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('>', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('<', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('>=', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('<=', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            ('!=', OperatorType.BINARY_LEFT_ASSOCIATIVE, 'not_equal'),
            ('=', OperatorType.BINARY_LEFT_ASSOCIATIVE, 'equal'),
            ('in', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('not', OperatorType.PREFIX_UNARY),
            (),
            ('and', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('or', OperatorType.BINARY_LEFT_ASSOCIATIVE),
            (),
            ('->', OperatorType.BINARY_RIGHT_ASSOCIATIVE),
        ]

    def insert_operator(self, existing_operator, existing_operator_binary,
                        new_operator, new_operator_type, create_group,
                        new_operator_alias=None):
        binary_types = (OperatorType.BINARY_RIGHT_ASSOCIATIVE,
                        OperatorType.BINARY_LEFT_ASSOCIATIVE)
        unary_types = (OperatorType.PREFIX_UNARY, OperatorType.SUFFIX_UNARY)
        position = 0
        if existing_operator is not None:
            position = -1
            for i, t in enumerate(self.operators):
                if len(t) < 2 or t[0] != existing_operator:
                    continue
                if existing_operator_binary and t[1] not in binary_types:
                    continue
                if not existing_operator_binary and t[1] not in unary_types:
                    continue
                position = i
                break
            if position < 0:
                raise ValueError('Operator {} is not found'.format(
                    existing_operator))
            while position < len(self.operators) and len(
                    self.operators[position]) > 1:
                position += 1
        if create_group:
            if position == len(self.operators):
                self.operators.append(())
                position += 1
            else:
                while position < len(self.operators) and len(
                        self.operators[position]) < 2:
                    position += 1
                self.operators.insert(position, ())
        self.operators.insert(
            position, (new_operator, new_operator_type, new_operator_alias))

    @staticmethod
    def _name_generator():
        value = 1
        while True:
            t = value
            chars = []
            while t:
                chars.append(chr(ord('A') + t % 26))
                t //= 26
            yield ''.join(chars)
            value += 1

    def _build_operator_table(self, name_generator):
        operators = {}
        name_value_op = None
        precedence = 1
        for record in self.operators:
            if not record:
                precedence += 1
                continue
            up, bp, name, alias = operators.get(record[0], (0, 0, '', None))
            if record[1] == OperatorType.NAME_VALUE_PAIR:
                if name_value_op is not None:
                    raise exceptions.InvalidOperatorTableException(record[0])
                name_value_op = record[0]
                continue
            if record[1] == OperatorType.PREFIX_UNARY:
                if up:
                    raise exceptions.InvalidOperatorTableException(record[0])
                up = precedence
            elif record[1] == OperatorType.SUFFIX_UNARY:
                if up:
                    raise exceptions.InvalidOperatorTableException(record[0])
                up = -precedence
            elif record[1] == OperatorType.BINARY_LEFT_ASSOCIATIVE:
                if bp:
                    raise exceptions.InvalidOperatorTableException(record[0])
                bp = precedence
            elif record[1] == OperatorType.BINARY_RIGHT_ASSOCIATIVE:
                if bp:
                    raise exceptions.InvalidOperatorTableException(record[0])
                bp = -precedence
            if record[0] == '[]':
                name = 'INDEXER'
            elif record[0] == '{}':
                name = 'MAP'
            else:
                name = name or 'OP_' + next(name_generator)
            operators[record[0]] = (
                up, bp, name, record[2] if len(record) > 2 else None)
        return YaqlOperators(operators, name_value_op)

    # noinspection PyMethodMayBeStatic
    def _create_lexer(self, operators):
        return lexer.Lexer(operators)

    # noinspection PyMethodMayBeStatic
    def _create_parser(self, lexer_rules, operators):
        return parser.Parser(lexer_rules, operators, self)

    def create(self, options=None):
        names = self._name_generator()
        operators = self._build_operator_table(names)
        lexer_rules = self._create_lexer(operators)
        ply_lexer = lex.lex(object=lexer_rules,
                            reflags=re.UNICODE | re.VERBOSE)
        ply_parser = yacc.yacc(
            module=self._create_parser(lexer_rules, operators),
            debug=False if not options else options.get('yaql.debug', False),
            tabmodule='m' + uuid.uuid4().hex, write_tables=False)

        return YaqlEngine(ply_lexer, ply_parser, options, self)
